import json
import os

import torch

from .. import utils
from .base import Base

__all__ = ["ColBERT"]


class ColBERT(Base):
    """ColBERT model.

    Parameters
    ----------
    model_name_or_path
        Path to the model or the model name.
    embedding_size
        Size of the embeddings in output of ColBERT model.
    device
        Device to use for the model. CPU or CUDA.
    kwargs
        Additional parameters to the SentenceTransformer model.

    Examples
    --------
    >>> from neural_cherche import models
    >>> import torch

    >>> _ = torch.manual_seed(42)

    >>> queries = [" ".join(["Berlin"] * 600), "Paris", "London"]

    >>> documents = [
    ...     "Berlin is the capital of Germany",
    ...     "Paris is the capital of France and France is in Europe",
    ...     "London is the capital of England",
    ... ]

    >>> encoder = models.ColBERT(
    ...     model_name_or_path="raphaelsty/neural-cherche-colbert",
    ...     device="cpu",
    ... )

    >>> scores = encoder.scores(
    ...    queries=queries,
    ...    documents=documents,
    ...    batch_size=10,
    ... )

    >>> scores
    tensor([19.2999,  3.6684,  3.5832])

    >>> _ = encoder.save_pretrained("checkpoint")

    >>> encoder = models.ColBERT(
    ...     model_name_or_path="checkpoint",
    ...     device="cpu",
    ... )

    >>> scores = encoder.scores(
    ...    queries=queries,
    ...    documents=documents,
    ...    batch_size=1,
    ... )

    >>> scores
    tensor([19.2999,  3.6684,  3.5832])

    >>> embeddings = encoder(
    ...     texts=queries,
    ...     query_mode=True
    ... )

    >>> embeddings["embeddings"].shape
    torch.Size([3, 32, 128])

    >>> embeddings = encoder(
    ...     texts=queries,
    ...     query_mode=False
    ... )

    >>> embeddings["embeddings"].shape
    torch.Size([3, 256, 128])

    """

    def __init__(
        self,
        model_name_or_path: str,
        embedding_size: int = 128,
        device: str = None,
        max_length_query: int = 64,
        max_length_document: int = 256,
        query_prefix: str = "[Q] ",
        document_prefix: str = "[D] ",
        padding: str = "max_length",
        truncation: bool = True,
        add_special_tokens: bool = True,
        freeze_layers_except_last_n: int = None,
        **kwargs,
    ) -> None:
        """Initialize the model."""
        super(ColBERT, self).__init__(
            model_name_or_path=model_name_or_path,
            device=device,
            extra_files_to_load=["linear.pt", "metadata.json"],
            query_prefix=query_prefix,
            document_prefix=document_prefix,
            padding=padding,
            truncation=truncation,
            add_special_tokens=add_special_tokens,
            freeze_layers_except_last_n=freeze_layers_except_last_n,
            **kwargs,
        )

        self.max_length_query = max_length_query
        self.max_length_document = max_length_document
        self.embedding_size = embedding_size

        if os.path.exists(path=os.path.join(self.model_folder, "linear.pt")):
            linear = torch.load(
                os.path.join(self.model_folder, "linear.pt"), map_location=self.device
            )
            self.embedding_size = linear["weight"].shape[0]
            in_features = linear["weight"].shape[1]
        else:
            with torch.no_grad():
                _, embeddings, _ = self._encode(texts=["test"])
                in_features = embeddings.shape[2]

        self.linear = torch.nn.Linear(
            in_features=in_features,
            out_features=self.embedding_size,
            dtype=torch.float32,
            bias=False,
            device=self.device,
        )

        torch.nn.init.xavier_uniform_(tensor=self.linear.weight)

        if os.path.exists(path=os.path.join(self.model_folder, "metadata.json")):
            with open(
                file=os.path.join(self.model_folder, "metadata.json"), mode="r"
            ) as f:
                metadata = json.load(fp=f)
            self.max_length_document = metadata["max_length_document"]
            self.max_length_query = metadata["max_length_query"]
            self.query_prefix = metadata.get("query_prefix", self.query_prefix)
            self.document_prefix = metadata.get("document_prefix", self.document_prefix)
            self.padding = metadata.get("padding", self.padding)
            self.truncation = metadata.get("truncation", self.truncation)
            self.add_special_tokens = metadata.get(
                "add_special_tokens", self.add_special_tokens
            )

        if os.path.exists(path=os.path.join(self.model_folder, "linear.pt")):
            self.linear.load_state_dict(state_dict=linear)

        self.query_pad_token = self.tokenizer.mask_token
        self.original_pad_token = self.tokenizer.pad_token

    def encode(
        self,
        texts: list[str],
        truncation: bool = True,
        add_special_tokens: bool = False,
        query_mode: bool = True,
        **kwargs,
    ) -> dict[str, torch.Tensor]:
        """Encode documents

        Parameters
        ----------
        texts
            List of sentences to encode.
        truncation
            Truncate the inputs.
        add_special_tokens
            Add special tokens.
        max_length
            Maximum length of the inputs.
        """
        with torch.no_grad():
            embeddings = self(
                texts=texts,
                query_mode=query_mode,
                **kwargs,
            )
        return embeddings

    def forward(
        self,
        texts: list[str],
        query_mode: bool = True,
        **kwargs,
    ) -> dict[str, torch.Tensor]:
        """Pytorch forward method.

        Parameters
        ----------
        texts
            List of sentences to encode.
        query_mode
            Wether to encode query or not.
        """
        prefix = self.query_prefix if query_mode else self.document_prefix
        texts = [prefix + text for text in texts]

        _, embeddings, attention_mask = self._encode(
            texts=texts,
            truncation=self.truncation,
            padding=self.padding,
            max_length=self.max_length_query
            if query_mode
            else self.max_length_document,
            add_special_tokens=self.add_special_tokens,
            **kwargs,
        )

        return {
            "embeddings": torch.nn.functional.normalize(
                input=self.linear(embeddings * attention_mask), p=2, dim=2
            )
        }

    def scores(
        self,
        queries: list[str],
        documents: list[str],
        batch_size: int = 2,
        tqdm_bar: bool = True,
        **kwargs,
    ) -> torch.Tensor:
        """Score queries and documents.

        Parameters
        ----------
        queries
            List of queries.
        documents
            List of documents.
        batch_size
            Batch size.
        truncation
            Truncate the inputs.
        add_special_tokens
            Add special tokens.
        tqdm_bar
            Show tqdm bar.
        """
        list_scores = []

        for batch_queries, batch_documents in zip(
            utils.batchify(
                X=queries,
                batch_size=batch_size,
                desc="Computing scores.",
                tqdm_bar=tqdm_bar,
            ),
            utils.batchify(X=documents, batch_size=batch_size, tqdm_bar=False),
        ):
            queries_embeddings = self.encode(
                texts=batch_queries,
                query_mode=True,
                **kwargs,
            )

            documents_embeddings = self.encode(
                texts=batch_documents,
                query_mode=False,
                **kwargs,
            )

            late_interactions = torch.einsum(
                "bsh,bth->bst",
                queries_embeddings["embeddings"],
                documents_embeddings["embeddings"],
            )

            late_interactions = torch.max(late_interactions, axis=2).values.sum(axis=1)

            list_scores.append(late_interactions)

        return torch.cat(tensors=list_scores, dim=0)

    def save_pretrained(self, path: str) -> "ColBERT":
        """Save model the model.

        Parameters
        ----------
        path
            Path to save the model.
        """
        self.model.save_pretrained(path)
        torch.save(self.linear.state_dict(), os.path.join(path, "linear.pt"))
        self.tokenizer.pad_token = self.original_pad_token
        self.tokenizer.save_pretrained(path)
        with open(os.path.join(path, "metadata.json"), "w") as f:
            json.dump(
                obj={
                    "max_length_query": self.max_length_query,
                    "max_length_document": self.max_length_document,
                    "query_prefix": self.query_prefix,
                    "document_prefix": self.document_prefix,
                    "padding": self.padding,
                    "truncation": self.truncation,
                    "add_special_tokens": self.add_special_tokens,
                },
                fp=f,
            )
        return self



import torch

__all__ = ["colbert_scores"]


def colbert_scores(
    anchor_embeddings: torch.Tensor,
    positive_embeddings: torch.Tensor,
    negative_embeddings: torch.Tensor,
    in_batch_negatives: bool = False,
) -> torch.Tensor:
    """ColBERT scoring function for training.

    Parameters
    ----------
    anchor_embeddings
        Anchor embeddings.
    positive_embeddings
        Positive embeddings.
    negative_embeddings
        Negative embeddings.
    in_batch_negatives
        Whether to use in batch negatives or not. Defaults to False.


    Examples
    --------
    >>> from neural_cherche import models, utils
    >>> import torch

    >>> _ = torch.manual_seed(42)

    >>> model = models.ColBERT(
    ...     model_name_or_path="raphaelsty/neural-cherche-colbert",
    ...     device="mps",
    ... )

    >>> anchor_embeddings = model(
    ...     ["Paris", "Toulouse"],
    ...     query_mode=True,
    ... )

    >>> positive_embeddings = model(
    ...    ["Paris", "Toulouse"],
    ...    query_mode=False,
    ... )

    >>> negative_embeddings = model(
    ...    ["Toulouse", "Paris"],
    ...    query_mode=False,
    ... )

    >>> scores = utils.colbert_scores(
    ...     anchor_embeddings=anchor_embeddings["embeddings"],
    ...     positive_embeddings=positive_embeddings["embeddings"],
    ...     negative_embeddings=negative_embeddings["embeddings"],
    ... )

    >>> scores
    {'positive_scores': tensor([24.7555, 26.4455], device='mps:0', grad_fn=<SumBackward1>), 'negative_scores': tensor([18.3089, 17.1017], device='mps:0', grad_fn=<SumBackward1>)}

    """
    positive_scores = (
        torch.einsum("bsh,bth->bst", anchor_embeddings, positive_embeddings)
        .max(axis=2)
        .values.sum(axis=1)
    )

    negative_scores = (
        torch.einsum("bsh,bth->bst", anchor_embeddings, negative_embeddings)
        .max(axis=2)
        .values.sum(axis=1)
    )

    if in_batch_negatives:
        in_batch_negative_scores = (
            torch.einsum(
                "bsh,bth->bst",
                anchor_embeddings,
                positive_embeddings.roll(shifts=1, dims=0),
            )
            .max(axis=2)
            .values.sum(axis=1)
        )

        negative_scores = torch.stack(
            tensors=[negative_scores, in_batch_negative_scores], axis=1
        )

    return {
        "positive_scores": positive_scores,
        "negative_scores": negative_scores,
    }



import numpy as np
from lenlp import sparse
from scipy.sparse import csc_matrix, csr_matrix, hstack, vstack

from .. import utils

__all__ = ["TfIdf"]


class TfIdf:
    """TfIdf retriever based on cosine similarities.

    Parameters
    ----------
    key
        Field identifier of each document.
    on
        Fields to use to match the query to the documents.
    documents
        Documents in TFIdf retriever are static. The retriever must be reseted to index new
        documents.
    k
        Number of documents to retrieve. Default is `None`, i.e all documents that match the query
        will be retrieved.
    tfidf
        TfidfVectorizer class of Sklearn to create a custom TfIdf retriever.

    Examples
    --------
    >>> from neural_cherche import retrieve
    >>> from pprint import pprint

    >>> documents = [
    ...     {"id": 0, "document": "Food"},
    ...     {"id": 1, "document": "Sports"},
    ...     {"id": 2, "document": "Cinema"},
    ... ]

    >>> queries = ["Food", "Sports", "Cinema"]

    >>> retriever = retrieve.TfIdf(
    ...     key="id",
    ...     on=["document"],
    ... )

    >>> documents_embeddings = retriever.encode_documents(
    ...     documents=documents
    ... )

    >>> retriever = retriever.add(
    ...     documents_embeddings=documents_embeddings,
    ... )

    >>> queries_embeddings = retriever.encode_queries(
    ...     queries=queries
    ... )

    >>> scores = retriever(
    ...     queries_embeddings=queries_embeddings,
    ...     k=4
    ... )

    >>> queries_embeddings = retriever.encode_queries(
    ...     queries=["hello world", "hello world"]
    ... )

    >>> pprint(scores)
    [[{'id': 0, 'similarity': 1.0}],
     [{'id': 1, 'similarity': 0.9999999999999999}],
     [{'id': 2, 'similarity': 0.9999999999999999}]]

    References
    ----------
    1. [sklearn.feature_extraction.text.TfidfVectorizer](https://scikit-learn.org/stable/modules/generated/sklearn.feature_extraction.text.TfidfVectorizer.html)
    2. [Python: tf-idf-cosine: to find document similarity](https://stackoverflow.com/questions/12118720/python-tf-idf-cosine-to-find-document-similarity)

    """

    def __init__(
        self,
        key: str,
        on: list[str],
        tfidf: sparse.TfidfVectorizer = None,
        fit: bool = True,
    ) -> None:
        self.key = key
        self.on = [on] if isinstance(on, str) else on

        self.vectorizer = (
            sparse.TfidfVectorizer(normalize=True, ngram_range=(3, 5), analyzer="char")
            if tfidf is None
            else tfidf
        )

        self.matrix = None
        self.fit = fit
        self.documents = []
        self.n_documents = 0

    def encode_documents(self, documents: list[dict]) -> dict[str, csr_matrix]:
        """Encode queries into sparse matrix.

        Parameters
        ----------
        documents
            Documents to encode.

        """
        content = [
            " ".join([doc.get(field, "") for field in self.on]) for doc in documents
        ]

        matrix = None
        if self.fit:
            matrix = self.vectorizer.fit_transform(raw_documents=content)
            self.fit = False

        # matrix is a csr matrix of shape (n_documents, n_features)
        if matrix is None:
            matrix = self.vectorizer.transform(raw_documents=content)
        return {document[self.key]: row for document, row in zip(documents, matrix)}

    def encode_queries(
        self, queries: list[str], warn_duplicates: bool = True
    ) -> dict[str, csr_matrix]:
        """Encode queries into sparse matrix.

        Parameters
        ----------
        queries
            Queries to encode.

        """
        if self.fit:
            raise ValueError("You must call the `encode_documents` method first.")

        # matrix is a csr matrix of shape (n_queries, n_features)
        matrix = self.vectorizer.transform(raw_documents=queries)
        embeddings = {query: row for query, row in zip(queries, matrix)}

        if len(embeddings) != len(queries) and warn_duplicates:
            utils.duplicates_queries_warning()

        return embeddings

    def add(
        self,
        documents_embeddings: dict[str, csr_matrix],
    ) -> "TfIdf":
        """Add new documents to the TFIDF retriever. The tfidf won't be refitted."""
        matrix = vstack(blocks=[row for row in documents_embeddings.values()]).T.tocsr()

        for document_key in documents_embeddings:
            self.documents.append({self.key: document_key})

        self.n_documents += len(documents_embeddings)
        self.matrix = (
            matrix if self.matrix is None else hstack(blocks=(self.matrix, matrix))
        )

        return self

    def top_k(self, similarities: csc_matrix, k: int) -> tuple[list, list]:
        """Return the top k documents for each query."""
        matchs, scores = [], []
        for row in similarities:
            _k = min(row.data.shape[0] - 1, k)
            ind = np.argpartition(a=row.data, kth=_k, axis=0)[:k]
            similarity = np.take_along_axis(arr=row.data, indices=ind, axis=0)
            indices = np.take_along_axis(arr=row.indices, indices=ind, axis=0)
            ind = np.argsort(a=similarity, axis=0)
            scores.append(-1 * np.take_along_axis(arr=similarity, indices=ind, axis=0))
            matchs.append(np.take_along_axis(arr=indices, indices=ind, axis=0))
        return matchs, scores

    def __call__(
        self,
        queries_embeddings: dict[str, csr_matrix],
        k: int = None,
        batch_size: int = 2000,
        tqdm_bar: bool = True,
    ) -> list[list[dict]]:
        """Retrieve documents from batch of queries.

        Parameters
        ----------
        queries_embeddings
            Queries embeddings.
        k
            Number of documents to retrieve. Default is `None`, i.e all documents that match the
            query will be retrieved.
        batch_size
            Batch size to use to retrieve documents.
        """
        k = k if k is not None else self.n_documents

        ranked = []

        for batch_embeddings in utils.batchify(
            list(queries_embeddings.values()),
            batch_size=batch_size,
            desc=f"{self.__class__.__name__} retriever",
            tqdm_bar=tqdm_bar,
        ):
            # self.matrix is a csr matrix of shape (n_features, n_documents)
            # Transform output a csr matrix of shape (n_queries, n_features)
            similarities = -1 * vstack(blocks=batch_embeddings).dot(self.matrix)
            batch_match, batch_similarities = self.top_k(similarities=similarities, k=k)

            for match, similarities in zip(batch_match, batch_similarities):
                ranked.append(
                    [
                        {**self.documents[idx], "similarity": similarity}
                        for idx, similarity in zip(match, similarities)
                        if similarity > 0
                    ]
                )

        return ranked



import torch

from .. import losses, utils

__all__ = ["train_colbert"]


def train_colbert(
    model,
    optimizer,
    anchor: list[str],
    positive: list[str],
    negative: list[str],
    in_batch_negatives: bool = False,
    step: int = None,
    gradient_accumulation_steps: int = 50,
    **kwargs,
) -> dict[str, torch.Tensor]:
    """Compute the ranking loss and the flops loss for a single step.

    Parameters
    ----------
    model
        Colbert model.
    optimizer
        Optimizer.
    anchor
        Anchor.
    positive
        Positive.
    negative
        Negative.
    in_batch_negatives
        Whether to use in batch negatives or not. Defaults to True.
    step
        Training step, if specified, will enable gradient_accumulation_steps.
    gradient_accumulation_steps
        Gradient accumulation steps. Defaults to 50.

    Examples
    --------
    >>> from neural_cherche import models, utils, train
    >>> import torch

    >>> _ = torch.manual_seed(42)

    >>> model = models.ColBERT(
    ...     model_name_or_path="raphaelsty/neural-cherche-colbert",
    ...     device="cpu",
    ... )

    >>> optimizer = torch.optim.AdamW(model.parameters(), lr=1e-2)

    >>> X = [
    ...     ("Sports", "Football", "Cinema"),
    ...     ("Sports", "Rugby", "Cinema"),
    ...     ("Sports", "Tennis", "Cinema"),
    ... ]

    >>> for step, (anchor, positive, negative) in enumerate(utils.iter(
    ...         X,
    ...         epochs=3,
    ...         batch_size=3,
    ...         shuffle=False
    ...     )):
    ...     loss = train.train_colbert(
    ...         model=model,
    ...         optimizer=optimizer,
    ...         anchor=anchor,
    ...         positive=positive,
    ...         negative=negative,
    ...         step=step,
    ...         gradient_accumulation_steps=2,
    ...         in_batch_negatives=True,
    ...     )

    >>> loss
    {'loss': tensor(1.0986, grad_fn=<NllLossBackward0>)}

    """

    anchor_embeddings = model(
        anchor,
        query_mode=True,
        **kwargs,
    )

    positive_embeddings = model(
        positive,
        query_mode=False,
        **kwargs,
    )

    negative_embeddings = model(
        negative,
        query_mode=False,
        **kwargs,
    )

    scores = utils.colbert_scores(
        anchor_embeddings=anchor_embeddings["embeddings"],
        positive_embeddings=positive_embeddings["embeddings"],
        negative_embeddings=negative_embeddings["embeddings"],
        in_batch_negatives=in_batch_negatives,
    )

    loss = losses.Ranking()(**scores)

    if step is not None:
        (loss / gradient_accumulation_steps).backward()

        if (step + 1) % gradient_accumulation_steps == 0:
            optimizer.step()
            optimizer.zero_grad(set_to_none=True)

    else:
        loss.backward()
        optimizer.step()
        optimizer.zero_grad(set_to_none=True)

    return {"loss": loss}



import torch

__all__ = ["dense_scores", "pairs_dense_scores"]


def _build_index(activations: torch.Tensor, embeddings: torch.Tensor) -> dict:
    """Build index to score documents using activated tokens and embeddings."""
    index = []
    for tokens_activation, tokens_embeddings in zip(activations, embeddings):
        index.append(
            {
                token.item(): embedding
                for token, embedding in zip(tokens_activation, tokens_embeddings)
            }
        )
    return index


def _intersection(t1: torch.Tensor, t2: torch.Tensor) -> list:
    """Retrieve intersection between two tensors."""
    t1, t2 = t1.flatten(), t2.flatten()
    combined = torch.cat((t1, t2), dim=0)
    uniques, counts = combined.unique(return_counts=True, sorted=False)
    return uniques[counts > 1].tolist()


def _get_intersection(queries_activations: list, documents_activations: list) -> list:
    """Retrieve intersection of activated tokens between queries and documents."""
    return [
        _intersection(query_activations, document_activations)
        for query_activations, document_activations in zip(
            queries_activations,
            documents_activations,
        )
    ]


def _get_scores(
    anchor_embeddings_index: dict,
    positive_embeddings_index: dict,
    negative_embeddings_index: dict,
    positive_intersections: torch.Tensor,
    negative_intersections: torch.Tensor,
    func,
) -> dict[str, torch.Tensor]:
    """Computes similarity scores between queries and documents based on activated tokens embeddings"""
    positive_scores, negative_scores = [], []

    for (
        anchor_embedding_index,
        positive_embedding_index,
        negative_embedding_index,
        positive_intersection,
        negative_intersection,
    ) in zip(
        anchor_embeddings_index,
        positive_embeddings_index,
        negative_embeddings_index,
        positive_intersections,
        negative_intersections,
    ):
        if len(positive_intersection) > 0 and len(negative_intersection) > 0:
            positive_scores.append(
                func(
                    torch.stack(
                        tensors=[
                            anchor_embedding_index[token]
                            for token in positive_intersection
                        ],
                        dim=0,
                    )
                    * torch.stack(
                        tensors=[
                            positive_embedding_index[token]
                            for token in positive_intersection
                        ],
                        dim=0,
                    )
                )
            )

            negative_scores.append(
                func(
                    torch.stack(
                        tensors=[
                            anchor_embedding_index[token]
                            for token in negative_intersection
                        ],
                        dim=0,
                    )
                    * torch.stack(
                        tensors=[
                            negative_embedding_index[token]
                            for token in negative_intersection
                        ],
                        dim=0,
                    )
                )
            )

    return {
        "positive_scores": torch.stack(
            positive_scores,
            dim=0,
        )
        if len(positive_scores) > 0
        else None,
        "negative_scores": torch.stack(
            negative_scores,
            dim=0,
        )
        if len(negative_scores) > 0
        else None,
    }


def dense_scores(
    anchor_activations: torch.Tensor,
    positive_activations: torch.Tensor,
    negative_activations: torch.Tensor,
    anchor_embeddings: torch.Tensor,
    positive_embeddings: torch.Tensor,
    negative_embeddings: torch.Tensor,
    func=torch.sum,
) -> dict[str, torch.Tensor]:
    """Computes score between queries and documents intersected activated tokens.

    Parameters
    ----------
    queries_activations
        Queries activated tokens.
    queries_embeddings
        Queries activated tokens embeddings.
    documents_activations
        Documents activated tokens.
    documents_embeddings
        Documents activated tokens embeddings.
    func
        Either torch.sum or torch.mean. torch.mean is dedicated to training and
        torch.sum is dedicated to inference.

    Examples
    --------
    >>> from neural_cherche import models, utils
    >>> import torch

    >>> _ = torch.manual_seed(42)

    >>> model = models.SparseEmbed(
    ...     model_name_or_path="raphaelsty/neural-cherche-sparse-embed",
    ...     device="mps",
    ... )

    >>> anchor_embeddings = model(
    ...     ["Paris", "Toulouse"],
    ...     query_mode=True,
    ... )

    >>> positive_embeddings = model(
    ...    ["Paris", "Toulouse"],
    ...    query_mode=False,
    ... )

    >>> negative_embeddings = model(
    ...    ["Toulouse", "Paris"],
    ...    query_mode=False,
    ... )

    >>> scores = utils.dense_scores(
    ...     anchor_activations=anchor_embeddings["activations"],
    ...     positive_activations=positive_embeddings["activations"],
    ...     negative_activations=negative_embeddings["activations"],
    ...     anchor_embeddings=anchor_embeddings["embeddings"],
    ...     positive_embeddings=positive_embeddings["embeddings"],
    ...     negative_embeddings=negative_embeddings["embeddings"],
    ...     func=torch.sum,
    ... )

    >>> scores
    {'positive_scores': tensor([144.4106, 155.5398], device='mps:0', grad_fn=<StackBackward0>), 'negative_scores': tensor([173.4966,  99.9521], device='mps:0', grad_fn=<StackBackward0>)}

    """
    anchor_embeddings_index = _build_index(
        activations=anchor_activations, embeddings=anchor_embeddings
    )

    positive_embeddings_index = _build_index(
        activations=positive_activations, embeddings=positive_embeddings
    )

    negative_embeddings_index = _build_index(
        activations=negative_activations, embeddings=negative_embeddings
    )

    positive_intersections = _get_intersection(
        queries_activations=anchor_activations,
        documents_activations=positive_activations,
    )

    negative_intersections = _get_intersection(
        queries_activations=anchor_activations,
        documents_activations=negative_activations,
    )

    return _get_scores(
        anchor_embeddings_index=anchor_embeddings_index,
        positive_embeddings_index=positive_embeddings_index,
        negative_embeddings_index=negative_embeddings_index,
        positive_intersections=positive_intersections,
        negative_intersections=negative_intersections,
        func=func,
    )


def pairs_dense_scores(
    queries_activations: torch.Tensor,
    documents_activations: torch.Tensor,
    queries_embeddings: torch.Tensor,
    documents_embeddings: torch.Tensor,
):
    """Scores pairs of queries and documents based on activated tokens."""
    queries_embeddings_index = _build_index(
        activations=queries_activations, embeddings=queries_embeddings
    )

    documents_embeddings_index = _build_index(
        activations=documents_activations, embeddings=documents_embeddings
    )

    intersections = _get_intersection(
        queries_activations=queries_activations,
        documents_activations=documents_activations,
    )

    scores = []

    for (
        query_embeddings_index,
        document_embedding_index,
        intersection,
    ) in zip(queries_embeddings_index, documents_embeddings_index, intersections):
        if len(intersection) > 0:
            scores.append(
                torch.sum(
                    torch.stack(
                        [query_embeddings_index[token] for token in intersection],
                        dim=0,
                    )
                    * torch.stack(
                        [document_embedding_index[token] for token in intersection],
                        dim=0,
                    )
                )
            )

        else:
            scores.append(torch.tensor(0.0, device=queries_activations.device))

    return torch.stack(
        scores,
        dim=0,
    )



import json
import os
import string

import torch

from .. import utils
from .base import Base

__all__ = ["Splade"]


class Splade(Base):
    """SpladeV1 model.

    Parameters
    ----------
    tokenizer
        HuggingFace Tokenizer.
    model
        HuggingFace AutoModelForMaskedLM.
    kwargs
        Additional parameters to the SentenceTransformer model.

    Examples
    --------
    >>> from neural_cherche import models
    >>> import torch

    >>> _ = torch.manual_seed(42)

    >>> model = models.Splade(
    ...     model_name_or_path="raphaelsty/neural-cherche-sparse-embed",
    ...     device="mps",
    ... )

    >>> queries_activations = model.encode(
    ...     ["Sports", "Music"],
    ... )

    >>> documents_activations = model.encode(
    ...    ["Music is great.", "Sports is great."],
    ...    query_mode=False,
    ... )

    >>> queries_activations["sparse_activations"].shape
    torch.Size([2, 30522])

    >>> model.scores(
    ...     queries=["Sports", "Music"],
    ...     documents=["Sports is great.", "Music is great."],
    ...     batch_size=2
    ... )
    tensor([14.4703, 13.9852], device='mps:0')

    >>> _ = model.save_pretrained("checkpoint")

    >>> model = models.Splade(
    ...     model_name_or_path="checkpoint",
    ...     device="mps",
    ... )

    >>> model.scores(
    ...     queries=["Sports", "Music"],
    ...     documents=["Sports is great.", "Music is great."],
    ...     batch_size=2
    ... )
    tensor([14.4703, 13.9852], device='mps:0')

    References
    ----------
    1. [SPLADE: Sparse Lexical and Expansion Model for First Stage Ranking](https://arxiv.org/abs/2107.05720)

    """

    def __init__(
        self,
        model_name_or_path: str = None,
        device: str = None,
        max_length_query: int = 64,
        max_length_document: int = 256,
        extra_files_to_load: list[str] = ["metadata.json"],
        query_prefix: str = "",
        document_prefix: str = "",
        padding: str = "max_length",
        truncation: bool | None = True,
        add_special_tokens: bool = True,
        n_mask_tokens: int = 3,
        freeze_layers_except_last_n: int = None,
        **kwargs,
    ) -> None:
        super(Splade, self).__init__(
            model_name_or_path=model_name_or_path,
            device=device,
            extra_files_to_load=extra_files_to_load,
            query_prefix=query_prefix,
            document_prefix=document_prefix,
            padding=padding,
            truncation=truncation,
            add_special_tokens=add_special_tokens,
            n_mask_tokens=n_mask_tokens,
            freeze_layers_except_last_n=freeze_layers_except_last_n,
            **kwargs,
        )

        self.relu = torch.nn.ReLU().to(self.device)

        if os.path.exists(path=os.path.join(self.model_folder, "metadata.json")):
            with open(
                file=os.path.join(self.model_folder, "metadata.json"), mode="r"
            ) as file:
                metadata = json.load(fp=file)

            max_length_query = metadata["max_length_query"]
            max_length_document = metadata["max_length_document"]
            self.query_prefix = metadata.get("query_prefix", self.query_prefix)
            self.document_prefix = metadata.get("document_prefix", self.document_prefix)
            self.padding = metadata.get("padding", self.padding)
            self.truncation = metadata.get("truncation", self.truncation)
            self.add_special_tokens = metadata.get(
                "add_special_tokens", self.add_special_tokens
            )
            self.n_mask_tokens = metadata.get("n_mask_tokens", self.n_mask_tokens)

        self.max_length_query = max_length_query
        self.max_length_document = max_length_document

    @torch.no_grad()
    def encode(
        self,
        texts: list[str],
        query_mode: bool = True,
        **kwargs,
    ) -> dict[str, torch.Tensor]:
        """Encode documents

        Parameters
        ----------
        texts
            List of documents to encode.
        truncation
            Whether to truncate the documents.
        padding
            Whether to pad the documents.
        max_length
            Maximum length of the documents.
        """
        return self(
            texts=texts,
            query_mode=query_mode,
            **kwargs,
        )

    def decode(
        self,
        sparse_activations: torch.Tensor,
        clean_up_tokenization_spaces: bool = False,
        skip_special_tokens: bool = True,
        k_tokens: int = 96,
    ) -> list[str]:
        """Decode activated tokens ids where activated value > 0.

        Parameters
        ----------
        sparse_activations
            Activated tokens.
        clean_up_tokenization_spaces
            Whether to clean up the tokenization spaces.
        skip_special_tokens
            Whether to skip special tokens.
        k_tokens
            Number of tokens to keep.
        """
        activations = self._filter_activations(
            sparse_activations=sparse_activations, k_tokens=k_tokens
        )

        # Decode
        return [
            " ".join(
                activation.translate(str.maketrans("", "", string.punctuation)).split()
            )
            for activation in self.tokenizer.batch_decode(
                sequences=activations,
                clean_up_tokenization_spaces=clean_up_tokenization_spaces,
                skip_special_tokens=skip_special_tokens,
            )
        ]

    def forward(
        self,
        texts: list[str],
        query_mode: bool,
        **kwargs,
    ) -> dict[str, torch.Tensor]:
        """Pytorch forward method.

        Parameters
        ----------
        texts
            List of documents to encode.
        query_mode
            Whether to encode queries or documents.
        """
        prefix = self.query_prefix if query_mode else self.document_prefix
        suffix = " ".join([self.tokenizer.mask_token] * self.n_mask_tokens)
        texts = [prefix + text + " " + suffix for text in texts]

        self.tokenizer.pad_token = (
            self.query_pad_token if query_mode else self.original_pad_token
        )

        k_tokens = self.max_length_query if query_mode else self.max_length_document

        logits, _, attention_mask = self._encode(
            texts=texts,
            truncation=self.truncation,
            padding=self.padding,
            add_special_tokens=self.add_special_tokens,
            max_length=self.max_length_query
            if query_mode
            else self.max_length_document,
            **kwargs,
        )

        activations = self._get_activation(logits=logits * attention_mask)

        activations = self._update_activations(
            **activations,
            k_tokens=k_tokens,
        )

        return {"sparse_activations": activations["sparse_activations"]}

    def save_pretrained(self, path: str) -> "Splade":
        """Save model the model.

        Parameters
        ----------
        path
            Path to save the model.

        """
        self.model.save_pretrained(path)
        self.tokenizer.pad_token = self.original_pad_token
        self.tokenizer.save_pretrained(save_directory=path)

        with open(file=os.path.join(path, "metadata.json"), mode="w") as file:
            json.dump(
                fp=file,
                obj={
                    "max_length_query": self.max_length_query,
                    "max_length_document": self.max_length_document,
                    "query_prefix": self.query_prefix,
                    "document_prefix": self.document_prefix,
                    "padding": self.padding,
                    "truncation": self.truncation,
                    "add_special_tokens": self.add_special_tokens,
                    "n_mask_tokens": self.n_mask_tokens,
                },
                indent=4,
            )

        return self

    def scores(
        self,
        queries: list[str],
        documents: list[str],
        batch_size: int = 32,
        tqdm_bar: bool = True,
        **kwargs,
    ) -> torch.Tensor:
        """Compute similarity scores between queries and documents.

        Parameters
        ----------
        queries
            List of queries.
        documents
            List of documents.
        batch_size
            Batch size.
        tqdm_bar
            Show a progress bar.
        """
        sparse_scores = []

        for batch_queries, batch_documents in zip(
            utils.batchify(
                X=queries,
                batch_size=batch_size,
                desc="Computing scores.",
                tqdm_bar=tqdm_bar,
            ),
            utils.batchify(X=documents, batch_size=batch_size, tqdm_bar=False),
        ):
            queries_embeddings = self.encode(
                texts=batch_queries,
                query_mode=True,
                **kwargs,
            )

            documents_embeddings = self.encode(
                texts=batch_documents,
                query_mode=False,
                **kwargs,
            )

            sparse_scores.append(
                torch.sum(
                    queries_embeddings["sparse_activations"]
                    * documents_embeddings["sparse_activations"],
                    axis=1,
                )
            )

        return torch.cat(tensors=sparse_scores, dim=0)

    def _get_activation(self, logits: torch.Tensor) -> dict[str, torch.Tensor]:
        """Returns activated tokens."""
        return {
            "sparse_activations": torch.amax(
                input=torch.log1p(input=self.relu(logits)), dim=1
            )
        }

    def _filter_activations(
        self, sparse_activations: torch.Tensor, k_tokens: int
    ) -> list[torch.Tensor]:
        """Among the set of activations, select the ones with a score > 0."""
        scores, activations = torch.topk(input=sparse_activations, k=k_tokens, dim=-1)
        return [
            torch.index_select(
                input=activation,
                dim=-1,
                index=torch.nonzero(input=score, as_tuple=True)[0],
            )
            for score, activation in zip(scores, activations)
        ]

    def _update_activations(
        self, sparse_activations: torch.Tensor, k_tokens: int
    ) -> torch.Tensor:
        """Returns activated tokens."""
        activations = torch.topk(input=sparse_activations, k=k_tokens, dim=1).indices

        # Set value of max sparse_activations which are not in top k to 0.
        sparse_activations = sparse_activations * torch.zeros(
            (sparse_activations.shape[0], sparse_activations.shape[1]),
            dtype=int,
            device=self.device,
        ).scatter_(dim=1, index=activations.long(), value=1)

        return {
            "activations": activations,
            "sparse_activations": sparse_activations,
        }



import torch

from .. import models, utils
from ..rank import ColBERT as ColBERTRanker

__all__ = ["ColBERT"]


class ColBERT(ColBERTRanker):
    """ColBERT retriever.

    Parameters
    ----------
    key
        Document unique identifier.
    on
        Document texts.
    model
        ColBERT model.
    device
        Device to use, default is model device.

    Examples
    --------
    >>> from neural_cherche import models, retrieve
    >>> from pprint import pprint
    >>> import torch

    >>> _ = torch.manual_seed(42)

    >>> encoder = models.ColBERT(
    ...     model_name_or_path="raphaelsty/neural-cherche-colbert",
    ...     device="mps",
    ... )

    >>> documents = [
    ...     {"id": 0, "document": "Food"},
    ...     {"id": 1, "document": "Sports"},
    ...     {"id": 2, "document": "Cinema"},
    ... ]

    >>> queries = ["Food", "Sports", "Cinema"]

    >>> retriever = retrieve.ColBERT(
    ...    key="id",
    ...    on=["document"],
    ...    model=encoder,
    ... )

    >>> documents_embeddings = retriever.encode_documents(
    ...     documents=documents,
    ...     batch_size=3,
    ... )

    >>> retriever = retriever.add(
    ...     documents_embeddings=documents_embeddings,
    ... )

    >>> queries_embeddings = retriever.encode_queries(
    ...     queries=queries,
    ...     batch_size=3,
    ... )

    >>> scores = retriever(
    ...     queries_embeddings=queries_embeddings,
    ...     batch_size=3,
    ...     tqdm_bar=True,
    ...     k=3,
    ... )

    >>> pprint(scores)
    [[{'id': 0, 'similarity': 4.7243194580078125},
      {'id': 2, 'similarity': 2.403003692626953},
      {'id': 1, 'similarity': 2.286036252975464}],
     [{'id': 1, 'similarity': 4.792296886444092},
      {'id': 2, 'similarity': 2.6001152992248535},
      {'id': 0, 'similarity': 2.312016487121582}],
     [{'id': 2, 'similarity': 5.069696426391602},
      {'id': 1, 'similarity': 2.5587477684020996},
      {'id': 0, 'similarity': 2.4474282264709473}]]

    """

    def __init__(
        self,
        key: str,
        on: list[str],
        model: models.ColBERT,
        device: str = None,
    ) -> None:
        self.key = key
        self.on = on if isinstance(on, list) else [on]
        self.model = model
        self.documents = []
        self.documents_embeddings = {}
        self.device = device if device is not None else self.model.device

    def add(
        self,
        documents_embeddings: dict[str, torch.Tensor],
    ) -> "ColBERT":
        """Add documents embeddings.

        Parameters
        ----------
        documents_embeddings
            Documents embeddings.
        documents
            Documents.
        """
        for document_key, tokens_embeddings in documents_embeddings.items():
            if document_key not in self.documents_embeddings:
                self.documents.append({self.key: document_key})
                self.documents_embeddings[document_key] = tokens_embeddings
        return self

    def __call__(
        self,
        queries_embeddings: dict[str, torch.Tensor],
        batch_size: int = 32,
        tqdm_bar: bool = True,
        k: int = None,
    ) -> list[list[str]]:
        """Rank documents  givent queries.

        Parameters
        ----------
        queries
            Queries.
        documents
            Documents.
        queries_embeddings
            Queries embeddings.
        documents_embeddings
            Documents embeddings.
        batch_size
            Batch size.
        tqdm_bar
            Show tqdm bar.
        k
            Number of documents to retrieve.
        """
        scores = []

        bar = utils.batchify(
            X=[query_embedding for query_embedding in queries_embeddings.values()],
            batch_size=1,
            tqdm_bar=tqdm_bar,
            desc=f"{self.__class__.__name__} retriever",
        )

        for query_embedding in bar:
            query_embedding = query_embedding[0]
            query_scores = []

            embedding_query = torch.tensor(
                data=query_embedding,
                device=self.device,
                dtype=torch.float32,
            )

            for batch_query_documents in utils.batchify(
                X=self.documents,
                batch_size=batch_size,
                tqdm_bar=False,
            ):
                embeddings_batch_documents = torch.stack(
                    tensors=[
                        torch.tensor(
                            data=self.documents_embeddings[document[self.key]],
                            device=self.device,
                            dtype=torch.float32,
                        )
                        for document in batch_query_documents
                    ],
                    dim=0,
                )

                query_documents_scores = torch.einsum(
                    "sh,bth->bst",
                    embedding_query,
                    embeddings_batch_documents,
                )

                query_scores.append(
                    query_documents_scores.max(dim=2).values.sum(axis=1)
                )

            scores.append(torch.cat(tensors=query_scores, dim=0))

        return self._rank(scores=scores, documents=self.documents, k=k)

    def _rank(
        self, scores: torch.Tensor, documents: list[list[dict]], k: int
    ) -> list[list[dict]]:
        """Rank documents by scores.

        Parameters
        ----------
        scores
            Scores.
        documents
            Documents.
        k
            Number of documents to retrieve.
        """
        ranked = []

        for query_scores in scores:
            top_k = torch.topk(
                input=query_scores,
                k=min(k, len(documents)) if k is not None else len(documents),
                dim=-1,
            )

            ranked.append(
                [
                    {**documents[indice], "similarity": similarity}
                    for indice, similarity in zip(top_k.indices, top_k.values.tolist())
                ]
            )

        return ranked



from .bm25 import BM25
from .colbert import ColBERT
from .sparse_embed import SparseEmbed
from .splade import Splade
from .tfidf import TfIdf

__all__ = [
    "ColBERT",
    "SparseEmbed",
    "Splade",
    "TfIdf",
    "BM25",
]



import torch

__all__ = ["sparse_scores"]


def sparse_scores(
    anchor_activations: torch.Tensor,
    positive_activations: torch.Tensor,
    negative_activations: torch.Tensor,
    in_batch_negatives: bool = False,
) -> dict[str, torch.Tensor]:
    """Computes dot product between anchor, positive and negative activations.

    Parameters
    ----------
    anchor_activations
        Activations of the anchors.
    positive_activations
        Activations of the positive documents.
    negative_activations
        Activations of the negative documents.
    in_batch_negatives
        Whether to use in batch negatives or not. Defaults to True.
        Sum up with negative scores the dot product.

    Examples
    --------
    >>> from neural_cherche import models

    >>> model = models.Splade(
    ...     model_name_or_path="raphaelsty/neural-cherche-sparse-embed",
    ...     device="mps"
    ... )

    >>> anchor_activations = model(
    ...     ["Sports", "Music"],
    ...     query_mode=True,
    ... )

    >>> positive_activations = model(
    ...    ["Sports", "Music"],
    ...    query_mode=False,
    ... )

    >>> negative_activations = model(
    ...    ["Cinema", "Movie"],
    ...    query_mode=False,
    ... )

    >>> sparse_scores(
    ...     anchor_activations=anchor_activations["sparse_activations"],
    ...     positive_activations=positive_activations["sparse_activations"],
    ...     negative_activations=negative_activations["sparse_activations"],
    ... )
    {'positive_scores': tensor([470.4049, 435.0986], device='mps:0', grad_fn=<SumBackward1>), 'negative_scores': tensor([301.5698, 353.6218], device='mps:0', grad_fn=<SumBackward1>)}

    """
    positive_scores = torch.sum(anchor_activations * positive_activations, axis=1)
    negative_scores = torch.sum(anchor_activations * negative_activations, axis=1)

    if in_batch_negatives:
        in_batch_negative_scores = torch.sum(
            anchor_activations * positive_activations.roll(shifts=1, dims=0), axis=1
        )

        negative_scores = torch.stack(
            tensors=[negative_scores, in_batch_negative_scores], axis=1
        )

    return {
        "positive_scores": positive_scores,
        "negative_scores": negative_scores,
    }



import os

from scipy.sparse import csr_matrix

from .. import models, utils
from .tfidf import TfIdf

__all__ = ["Splade"]


class Splade(TfIdf):
    """Retriever class.

    Parameters
    ----------
    key
        Document unique identifier.
    on
        Document texts.
    model
        SparsEmbed model.

    Examples
    --------
    >>> from neural_cherche import models, retrieve
    >>> from pprint import pprint
    >>> import torch

    >>> _ = torch.manual_seed(42)

    >>> documents = [
    ...     {"id": 0, "document": "Food"},
    ...     {"id": 1, "document": "Sports"},
    ...     {"id": 2, "document": "Cinema"},
    ... ]

    >>> queries = ["Food", "Sports", "Cinema"]

    >>> model = models.Splade(
    ...     model_name_or_path="raphaelsty/neural-cherche-sparse-embed",
    ...     device="mps",
    ... )

    >>> retriever = retrieve.Splade(
    ...     key="id",
    ...     on="document",
    ...     model=model
    ... )

    >>> documents_embeddings = retriever.encode_documents(
    ...     documents=documents,
    ...     batch_size=32,
    ... )

    >>> queries_embeddings = retriever.encode_queries(
    ...     queries=queries,
    ...     batch_size=32,
    ... )

    >>> retriever = retriever.add(
    ...     documents_embeddings=documents_embeddings,
    ... )

    >>> scores = retriever(
    ...     queries_embeddings=queries_embeddings,
    ...     k=3,
    ... )

    >>> pprint(scores)
    [[{'id': 0, 'similarity': 19.365147},
      {'id': 1, 'similarity': 4.440977},
      {'id': 2, 'similarity': 3.9301493}],
     [{'id': 1, 'similarity': 18.055237},
      {'id': 0, 'similarity': 4.440977},
      {'id': 2, 'similarity': 3.6559556}],
     [{'id': 2, 'similarity': 16.21696},
      {'id': 0, 'similarity': 3.9301493},
      {'id': 1, 'similarity': 3.6559556}]]

    """

    def __init__(
        self,
        key: str,
        on: list[str],
        model: models.Splade,
        tokenizer_parallelism: str = "false",
    ) -> None:
        super().__init__(
            key=key,
            on=on,
        )

        self.model = model
        os.environ["TOKENIZERS_PARALLELISM"] = tokenizer_parallelism

    def encode_documents(
        self,
        documents: list[dict],
        batch_size: int = 32,
        tqdm_bar: bool = True,
        query_mode: bool = False,
        desc: str = "documents embeddings",
        **kwargs,
    ) -> dict[str, csr_matrix]:
        """Encode queries into sparse matrix.

        Parameters
        ----------
        documents
            Documents to encode.
        batch_size
            Batch size.
        tqdm_bar
            Whether to show tqdm bar.

        """
        documents_embeddings = {}

        for batch_documents in utils.batchify(
            documents,
            batch_size=batch_size,
            tqdm_bar=tqdm_bar,
            desc=f"{self.__class__.__name__} {desc}",
        ):
            embeddings = self.model.encode(
                [
                    " ".join([doc.get(field, "") for field in self.on])
                    for doc in batch_documents
                ],
                query_mode=query_mode,
                **kwargs,
            )

            sparse_activations = csr_matrix(
                embeddings["sparse_activations"].detach().cpu()
            )
            for document, sparse_activation in zip(batch_documents, sparse_activations):
                documents_embeddings[document[self.key]] = sparse_activation

        return documents_embeddings

    def encode_queries(
        self,
        queries: list[str],
        batch_size: int = 32,
        tqdm_bar: bool = True,
        query_mode: bool = True,
        desc: str = "queries embeddings",
        warn_duplicates: bool = True,
        **kwargs,
    ) -> dict[str, csr_matrix]:
        """Encode queries into sparse matrix.

        Parameters
        ----------
        documents
            Documents to encode.
        batch_size
            Batch size.
        tqdm_bar
            Whether to show tqdm bar.

        """
        queries_embeddings = {}

        for batch_queries in utils.batchify(
            queries,
            batch_size=batch_size,
            tqdm_bar=tqdm_bar,
            desc=f"{self.__class__.__name__}  {desc}",
        ):
            embeddings = self.model.encode(
                batch_queries,
                query_mode=query_mode,
                **kwargs,
            )

            sparse_activations = csr_matrix(
                embeddings["sparse_activations"].detach().cpu()
            )
            for query, sparse_activation in zip(batch_queries, sparse_activations):
                queries_embeddings[query] = sparse_activation

        if len(embeddings) != len(queries) and warn_duplicates:
            utils.duplicates_queries_warning()

        return queries_embeddings



import collections
import os

import numpy as np
import torch
from scipy.sparse import csr_matrix, hstack, vstack

from .. import models, utils
from .tfidf import TfIdf

__all__ = ["SparseEmbed"]


class SparseEmbed(TfIdf):
    """SparseEmbed retriever.

    Parameters
    ----------
    key
        Document unique identifier.
    on
        Document texts.
    model
        SparseEmbed model.

    Examples
    --------
    >>> from neural_cherche import models, retrieve
    >>> from pprint import pprint
    >>> import torch

    >>> _ = torch.manual_seed(42)

    >>> model = models.SparseEmbed(
    ...     model_name_or_path="raphaelsty/neural-cherche-sparse-embed",
    ...     device="cpu",
    ...     embedding_size=64,
    ... )

    >>> retriever = retrieve.SparseEmbed(
    ...     key="id",
    ...     on="document",
    ...     model=model,
    ... )

    >>> documents = [
    ...     {"id": 0, "document": " ".join(["Food Hello world"] * 200)},
    ...     {"id": 1, "document": "Sports"},
    ...     {"id": 2, "document": "Cinema"},
    ... ]

    >>> queries = ["Food", "Sports", "Cinema"]

    >>> documents_embeddings = retriever.encode_documents(
    ...     documents=documents,
    ...     batch_size=2,
    ... )

    >>> queries_embeddings = retriever.encode_queries(
    ...     queries=queries,
    ...     batch_size=2,
    ... )

    >>> retriever = retriever.add(
    ...     documents_embeddings=documents_embeddings,
    ... )

    >>> scores = retriever(
    ...     queries_embeddings=queries_embeddings,
    ...     batch_size=32
    ... )

    >>> pprint(scores)
    [[{'id': 0, 'similarity': 650.8814697265625},
      {'id': 2, 'similarity': 63.36189270019531},
      {'id': 1, 'similarity': 55.061676025390625}],
     [{'id': 1, 'similarity': 139.74667358398438},
      {'id': 2, 'similarity': 63.8810920715332},
      {'id': 0, 'similarity': 42.96449661254883}],
     [{'id': 2, 'similarity': 158.86715698242188},
      {'id': 0, 'similarity': 68.95826721191406},
      {'id': 1, 'similarity': 63.872554779052734}]]

    """

    def __init__(
        self,
        key: str,
        on: list[str],
        model: models.SparseEmbed,
        tokenizer_parallelism: str = "false",
        device: str = None,
    ) -> None:
        super().__init__(
            key=key,
            on=on,
        )

        self.model = model

        # TfIdf Retriever.
        self.fit = True

        # Documents embeddings and activations store.
        self.documents_embeddings, self.documents_activations = [], []

        os.environ["TOKENIZERS_PARALLELISM"] = tokenizer_parallelism
        self.device = device if device is not None else self.model.device

    def encode_queries(
        self,
        queries: list[str],
        batch_size: int = 32,
        tqdm_bar: bool = True,
        query_mode: bool = True,
        warn_duplicates: bool = True,
        desc: str = "embeddings queries",
        **kwargs,
    ) -> dict[str, dict[str, torch.Tensor]]:
        """Encode queries.

        Parameters
        ----------
        queries
            List of queries to encode.
        batch_size
            Batch size.
        tqdm_bar
            Display a tqdm bar.
        """
        embeddings = collections.defaultdict(dict)

        for batch in utils.batchify(
            queries,
            batch_size=batch_size,
            desc=f"{self.__class__.__name__} {desc}",
            tqdm_bar=tqdm_bar,
        ):
            queries_embeddings = self.model.encode(
                texts=batch,
                query_mode=query_mode,
                **kwargs,
            )

            queries_embeddings = {
                field: array.detach().cpu().numpy().astype(np.float32)
                for field, array in queries_embeddings.items()
            }

            queries_embeddings["sparse_activations"] = csr_matrix(
                queries_embeddings["sparse_activations"]
            )

            for query, sparse_activations, activations, tokens_embeddings in zip(
                batch,
                queries_embeddings["sparse_activations"],
                queries_embeddings["activations"],
                queries_embeddings["embeddings"],
            ):
                embeddings[query]["sparse_activations"] = csr_matrix(sparse_activations)
                embeddings[query]["activations"] = activations
                embeddings[query]["embeddings"] = tokens_embeddings

        if len(embeddings) != len(queries) and warn_duplicates:
            utils.duplicates_queries_warning()

        return embeddings

    def encode_documents(
        self,
        documents: list[dict],
        batch_size: int = 32,
        tqdm_bar: bool = True,
        query_mode: bool = False,
        desc: str = "embeddings documents",
        **kwargs,
    ) -> dict[str, dict[str, torch.Tensor]]:
        """Encode documents.

        Parameters
        ----------
        documents
            List of documents to encode.
        """
        embeddings = collections.defaultdict(dict)

        for batch in utils.batchify(
            X=documents,
            batch_size=batch_size,
            tqdm_bar=tqdm_bar,
            desc=f"{self.__class__.__name__} {desc}",
        ):
            documents_embeddings = self.model.encode(
                texts=[
                    " ".join([doc.get(field, "") for field in self.on]) for doc in batch
                ],
                query_mode=query_mode,
                **kwargs,
            )

            documents_embeddings = {
                field: array.detach().cpu().numpy().astype(np.float32)
                for field, array in documents_embeddings.items()
            }

            documents_embeddings["sparse_activations"] = csr_matrix(
                documents_embeddings["sparse_activations"]
            )

            for document, sparse_activations, activations, tokens_embeddings in zip(
                batch,
                documents_embeddings["sparse_activations"],
                documents_embeddings["activations"],
                documents_embeddings["embeddings"],
            ):
                key = document[self.key]
                embeddings[key]["sparse_activations"] = sparse_activations
                embeddings[key]["activations"] = activations
                embeddings[key]["embeddings"] = tokens_embeddings

        return embeddings

    def add(
        self,
        documents_embeddings: dict[dict[str, torch.Tensor]],
    ) -> "SparseEmbed":
        """Add documents embeddings and activations to the retriever.

        Parameters
        ----------
        documents
            Documents to add.
        documents_embeddings
            Documents embeddings.
        build_index
            Build a sparse matrix index.
        matrix
            List of sparse matrix.
        """
        matrix = vstack(
            [
                embeddings["sparse_activations"]
                for embeddings in documents_embeddings.values()
            ]
        ).T.tocsr()

        self.matrix = matrix if self.matrix is None else hstack((self.matrix, matrix))

        for key, embeddings in documents_embeddings.items():
            self.documents_embeddings.append(
                {
                    token: token_embedding
                    for token, token_embedding in zip(
                        embeddings["activations"],
                        embeddings["embeddings"],
                    )
                }
            )
            self.documents_activations.append(embeddings["activations"])

        for key in documents_embeddings:
            self.documents.append({self.key: key})
            self.n_documents += 1

        return self

    def __call__(
        self,
        queries_embeddings: dict[str, dict[str, torch.Tensor]],
        k: int = None,
        k_rank: int = None,
        batch_size: int = 64,
        tqdm_bar: bool = True,
    ) -> list:
        """Retrieve documents.

        Parameters
        ----------
        q
            Queries.
        k
            Number of documents to retrieve.
        batch_size
            Batch size.
        tqdm_bar
            Display a tqdm bar.
        """
        k = k if k is not None else self.n_documents
        k_rank = k_rank if k_rank is not None else k

        ranked = []

        for queries_batch in utils.batchify(
            X=list(queries_embeddings.values()),
            batch_size=batch_size,
            desc=f"{self.__class__.__name__} retriever",
            tqdm_bar=tqdm_bar,
        ):
            embeddings = {
                "activations": torch.stack(
                    tensors=[
                        torch.tensor(data=query["activations"], device=self.device)
                        for query in queries_batch
                    ],
                    dim=0,
                ),
                "embeddings": torch.stack(
                    tensors=[
                        torch.tensor(data=query["embeddings"], device=self.device)
                        for query in queries_batch
                    ],
                    dim=0,
                ),
                "sparse_activations": vstack(
                    blocks=[query["sparse_activations"] for query in queries_batch]
                ),
            }

            ranked.extend(
                self._retrieve(
                    embeddings=embeddings,
                    k=k,
                    k_rank=k_rank,
                )
            )

        return ranked

    def _retrieve(
        self, embeddings: dict[str, torch.Tensor], k: int, k_rank: int
    ) -> list[list[dict]]:
        """Retrieve documents from input embeddings.

        Parameters
        ----------
        embeddings
            Input embeddings.
        k
            Number of documents to retrieve.
        """
        sparse_matchs = [
            activations.indices
            for activations in (embeddings["sparse_activations"].dot(self.matrix))
        ]

        documents_activations = [
            [
                torch.tensor(
                    data=self.documents_activations[document], device=self.device
                )
                for document in query_matchs
            ]
            for query_matchs in sparse_matchs
        ]

        intersections = self._get_intersection(
            queries_activations=embeddings["activations"],
            documents_activations=documents_activations,
        )

        queries_embeddings = [
            {
                token: token_embedding
                for token, token_embedding in zip(query_activations, query_embeddings)
            }
            for query_activations, query_embeddings in zip(
                embeddings["activations"].tolist(),
                embeddings["embeddings"],
            )
        ]

        dense_scores = self._get_scores(
            queries_embeddings=queries_embeddings,
            documents_embeddings=[
                [self.documents_embeddings[document] for document in match]
                for match in sparse_matchs
            ],
            intersections=intersections,
        )

        return self._rank(
            dense_scores=dense_scores,
            sparse_matchs=sparse_matchs,
            k=k_rank,
        )

    def _rank(
        self, dense_scores: torch.Tensor, sparse_matchs: torch.Tensor, k: int
    ) -> list[dict]:
        """Rank documents by scores.

        Parameters
        ----------
        scores
            Scores between queries and documents.
        matchs
            Documents matchs.
        """
        ranked = []

        for query_scores, query_sparse_matchs in zip(dense_scores, sparse_matchs):
            query_scores, query_matchs = torch.topk(
                input=query_scores,
                k=min(k, query_scores.shape[0]),
                dim=-1,
            )

            if len(query_scores) == 0:
                ranked.append([])
                continue

            query_matchs = np.take(
                a=query_sparse_matchs, indices=query_matchs.cpu().detach().numpy()
            )

            query_scores = query_scores.tolist()

            ranked.append(
                [
                    {
                        **self.documents[document],
                        "similarity": score,
                    }
                    for score, document in zip(query_scores, query_matchs)
                ]
            )

        return ranked

    @classmethod
    def _get_intersection(
        cls,
        queries_activations: list[torch.Tensor],
        documents_activations: list[list[torch.Tensor]],
    ) -> list[list[torch.Tensor]]:
        """Retrieve intersection of activated tokens between queries and documents."""
        return [
            [
                cls._intersection(t1=query_activations, t2=document_activations)
                for document_activations in query_documents_activations
            ]
            for query_activations, query_documents_activations in zip(
                queries_activations, documents_activations
            )
        ]

    @staticmethod
    def _intersection(t1: torch.Tensor, t2: torch.Tensor) -> list[int]:
        t1, t2 = t1.flatten(), t2.flatten()
        combined = torch.cat(tensors=(t1, t2), dim=0)
        uniques, counts = combined.unique(return_counts=True, sorted=False)
        return uniques[counts > 1].tolist()

    def _get_scores(
        self,
        queries_embeddings: list[dict[int, torch.Tensor]],
        documents_embeddings: list[list[dict[int, torch.Tensor]]],
        intersections: list[torch.Tensor],
    ) -> list[torch.Tensor]:
        """Computes similarity scores between queries and documents with activated tokens embeddings.

        Parameters
        ----------
        queries_embeddings
            List of dictionaries, with tokens as keys and embeddings as values.
        documents_embeddings
            List of list of dictionnaries with tokens as keys and embeddings as values.

        """
        queries_documents_scores = []

        for intersection, query_embeddings, document_embeddings in zip(
            intersections, queries_embeddings, documents_embeddings
        ):
            query_documents_scores = []

            for intersection, document_embedding in zip(
                intersection, document_embeddings
            ):
                if len(intersection) > 0:
                    query_documents_scores.append(
                        torch.sum(
                            input=torch.stack(
                                tensors=[
                                    torch.tensor(
                                        data=document_embedding[token],
                                        device=self.device,
                                    )
                                    for token in intersection
                                ],
                                dim=0,
                            )
                            * torch.stack(
                                tensors=[
                                    query_embeddings[token] for token in intersection
                                ],
                                dim=0,
                            )
                        )
                    )

                else:
                    query_documents_scores.append(
                        torch.tensor(data=0.0, device=self.device)
                    )

            if not query_documents_scores:
                queries_documents_scores.append([])
                continue

            queries_documents_scores.append(
                torch.stack(tensors=query_documents_scores, dim=0)
            )

        return queries_documents_scores



import numpy as np
from lenlp import sparse
from scipy.sparse import csr_matrix, hstack, vstack
from sklearn.utils.sparsefuncs_fast import inplace_csr_row_normalize_l2

from .tfidf import TfIdf

__all__ = ["BM25"]


class BM25(TfIdf):
    """BM25 retriever.

    Parameters
    ----------
    key
        Field identifier of each document.
    on
        Fields to use to match the query to the documents.
    documents
        Documents in TFIdf retriever are static. The retriever must be reseted to index new
        documents.
    CountVectorizer
        CountVectorizer class of Sklearn to create a custom CountVectorizer counter.
    b
        The impact of document length normalization.  Default is `0.75`, Higher will
        penalize longer documents more.
    k1
        How quickly the impact of term frequency saturates.  Default is `1.5`, Higher
        will make term frequency more influential.
    epsilon
        Smoothing term. Default is `0`.
    fit
        Fit the CountVectorizer on the documents. Default is `True`.

    Examples
    --------
    >>> from neural_cherche import retrieve
    >>> from pprint import pprint

    >>> documents = [
    ...     {"id": 0, "document": "Food"},
    ...     {"id": 1, "document": "Sports"},
    ...     {"id": 2, "document": "Cinema"},
    ... ]

    >>> queries = ["Food", "Sports", "Cinema food sports", "cinema"]

    >>> retriever = retrieve.BM25(
    ...     key="id",
    ...     on=["document"],
    ... )

    >>> documents_embeddings = retriever.encode_documents(
    ...     documents=documents
    ... )

    >>> retriever = retriever.add(
    ...     documents_embeddings=documents_embeddings,
    ... )

    >>> queries_embeddings = retriever.encode_queries(
    ...     queries=queries
    ... )

    >>> scores = retriever(
    ...     queries_embeddings=queries_embeddings,
    ...     k=4
    ... )

    >>> pprint(scores)
    [[{'id': 0, 'similarity': 3.0}],
     [{'id': 1, 'similarity': 9.0}],
     [{'id': 2, 'similarity': 9.0},
      {'id': 1, 'similarity': 9.0},
      {'id': 0, 'similarity': 3.0}],
     [{'id': 2, 'similarity': 9.0}]]

    >>> documents = [
    ...     {"id": 3, "document": "Food"},
    ...     {"id": 4, "document": "Sports"},
    ...     {"id": 5, "document": "Cinema"},
    ... ]

    >>> documents_embeddings = retriever.encode_documents(
    ...     documents=documents
    ... )

    >>> retriever = retriever.add(
    ...     documents_embeddings=documents_embeddings,
    ... )

    >>> scores = retriever(
    ...     queries_embeddings=queries_embeddings,
    ...     k=4
    ... )

    >>> pprint(scores)
    [[{'id': 3, 'similarity': 2.432886242866516},
      {'id': 0, 'similarity': 1.7552960515022278}],
     [{'id': 1, 'similarity': 6.648760557174683},
      {'id': 4, 'similarity': 6.065804421901703}],
     [{'id': 1, 'similarity': 6.648760557174683},
      {'id': 2, 'similarity': 6.648760557174683},
      {'id': 4, 'similarity': 6.065804421901703},
      {'id': 5, 'similarity': 6.065804421901703}],
     [{'id': 2, 'similarity': 6.648760557174683},
      {'id': 5, 'similarity': 6.065804421901703}]]


    References
    ----------
    1. [sklearn.feature_extraction.text.CountVectorizer](https://scikit-learn.org/stable/modules/generated/sklearn.feature_extraction.text.CountVectorizer.html)
    2. [Python: tf-idf-cosine: to find document similarity](https://stackoverflow.com/questions/12118720/python-tf-idf-cosine-to-find-document-similarity)

    """

    def __init__(
        self,
        key: str,
        on: list[str],
        count_vectorizer: sparse.CountVectorizer = None,
        b: float = 0.75,
        k1: float = 1.5,
        epsilon: float = 0,
        fit: bool = True,
    ) -> None:
        super().__init__(
            key=key,
            on=on,
            tfidf=(
                sparse.CountVectorizer(
                    normalize=True, ngram_range=(3, 5), analyzer="char_wb"
                )
                if count_vectorizer is None
                else count_vectorizer
            ),
            fit=fit,
        )

        self.b = b
        self.k1 = k1
        self.epsilon = epsilon
        self.tf = None

    def add(
        self,
        documents_embeddings: dict[str, csr_matrix],
    ) -> "BM25":
        """Add new documents to the TFIDF retriever. The tfidf won't be refitted."""
        matrix = vstack(
            blocks=[row for row in documents_embeddings.values()], dtype=np.float32
        )

        self.tf = (
            matrix.sum(axis=0) if self.tf is None else self.tf + matrix.sum(axis=0)
        )

        len_documents = (matrix).sum(axis=1)

        average_len_documents = len_documents.mean()

        regularization = np.squeeze(
            np.asarray(
                (
                    self.k1
                    * (1 - self.b + self.b * (len_documents / average_len_documents))
                ).flatten()
            )
        )

        numerator = matrix.copy()
        denominator = matrix.copy().tocsc()

        numerator.data = numerator.data * (self.k1 + 1)
        denominator.data += np.take(a=regularization, indices=denominator.indices)
        matrix.data = (numerator.data / denominator.tocsr().data) + self.epsilon

        for document_key in documents_embeddings:
            self.documents.append({self.key: document_key})

        self.n_documents += len(documents_embeddings)

        idf = np.squeeze(
            a=np.asarray(
                a=np.log((self.n_documents - self.tf + 0.5) / (self.tf + 0.5) + 1)
            )
        )

        matrix = matrix.multiply(idf).T.tocsr()

        self.matrix = (
            matrix if self.matrix is None else hstack(blocks=(self.matrix, matrix))
        )

        inplace_csr_row_normalize_l2(X=self.matrix)
        return self



import warnings

__all__ = ["duplicates_queries_warning"]


def duplicates_queries_warning() -> None:
    message = """Duplicate queries found. Provide distinct queries."""
    warnings.warn(message=message)



import torch

__all__ = ["Flops", "FlopsScheduler"]


class FlopsScheduler:
    """Flops scheduler.

    References
    ----------
    1. [MINIMIZING FLOPS TO LEARN EFFICIENT SPARSE REPRESENTATIONS](https://arxiv.org/pdf/2004.05665.pdf)
    2. [SPLADE: Sparse Lexical and Expansion Model for First Stage Ranking](https://arxiv.org/pdf/2107.05720.pdf)
    """

    def __init__(self, weight: float = 3e-6, steps: int = 10000):
        self._weight = weight
        self.weight = 0.0
        self.steps = steps
        self._step = 1

    def get(self) -> float:
        if self._step >= self.steps:
            pass
        else:
            self._step += 1
            self.weight = self._weight * (self._step / self.steps) ** 2
        return self.weight


class Flops(torch.nn.Module):
    """Flops loss, act as regularization loss over sparse activations.

    Examples
    --------
    >>> from neural_cherche import models, utils, losses
    >>> import torch

    >>> _ = torch.manual_seed(42)

    >>> model = models.Splade(
    ...     model_name_or_path="raphaelsty/neural-cherche-sparse-embed",
    ...     device="cpu",
    ... )

    >>> anchor_activations = model(
    ...     ["Sports", "Music"],
    ...     query_mode=True,
    ... )

    >>> positive_activations = model(
    ...    ["Sports", "Music"],
    ...     query_mode=False,
    ... )

    >>> negative_activations = model(
    ...    ["Cinema", "Movie"],
    ...     query_mode=False,
    ... )

    >>> losses.Flops()(
    ...     anchor_activations=anchor_activations["sparse_activations"],
    ...     positive_activations=positive_activations["sparse_activations"],
    ...     negative_activations=negative_activations["sparse_activations"],
    ... )
    tensor(1., grad_fn=<ClampBackward1>)

    References
    ----------
    1. [MINIMIZING FLOPS TO LEARN EFFICIENT SPARSE REPRESENTATIONS](https://arxiv.org/pdf/2004.05665.pdf)
    2. [SPLADE: Sparse Lexical and Expansion Model for First Stage Ranking](https://arxiv.org/pdf/2107.05720.pdf)

    """

    def __init__(self) -> None:
        super(Flops, self).__init__()

    def __call__(
        self,
        anchor_activations: torch.Tensor,
        positive_activations: torch.Tensor,
        negative_activations: torch.Tensor,
        threshold: float = 30.0,
        max_flops_loss: float = 1.0,
    ) -> torch.Tensor:
        """Loss which tend to reduce sparse activation."""
        activations = torch.cat(
            tensors=[anchor_activations, positive_activations, negative_activations],
            dim=0,
        )

        return torch.abs(
            input=threshold
            - torch.sum(
                input=torch.mean(input=torch.abs(input=activations), dim=0) ** 2, dim=0
            )
        ).clip(min=0.0, max=max_flops_loss)



import torch

__all__ = ["Ranking"]


class Ranking(torch.nn.Module):
    """Ranking loss.

    Examples
    --------
    >>> from neural_cherche import models, utils, losses
    >>> from pprint import pprint as print

    >>> model = models.Splade(
    ...     model_name_or_path="raphaelsty/splade-max",
    ...     device="cpu",
    ... )

    >>> queries_activations = model(
    ...     ["Sports", "Music"],
    ...     query_mode=True,
    ... )

    >>> positive_activations = model(
    ...    ["Sports", "Music"],
    ...    query_mode=False,
    ... )

    >>> negative_activations = model(
    ...    ["Cinema", "Movie"],
    ...    query_mode=False,
    ... )

    >>> scores = utils.sparse_scores(
    ...     anchor_activations=queries_activations["sparse_activations"],
    ...     positive_activations=positive_activations["sparse_activations"],
    ...     negative_activations=negative_activations["sparse_activations"],
    ...     in_batch_negatives=True,
    ... )

    >>> losses.Ranking()(**scores)
    tensor(6.8423e-05, grad_fn=<NllLossBackward0>)

    References
    ----------
    1. [SPLADE: Sparse Lexical and Expansion Model for First Stage Ranking](https://arxiv.org/pdf/2107.05720.pdf)

    """

    def __init__(self) -> None:
        super(Ranking, self).__init__()
        self.cross_entropy = torch.nn.CrossEntropyLoss()

    def __call__(
        self,
        positive_scores: torch.Tensor,
        negative_scores: torch.Tensor,
    ) -> torch.Tensor:
        """Ranking loss."""
        positive_scores = positive_scores.unsqueeze(dim=-1)
        if negative_scores.ndim == 1:
            negative_scores = negative_scores.unsqueeze(dim=-1)

        scores = torch.cat(
            tensors=[
                positive_scores,
                negative_scores,
            ],
            dim=-1,
        )

        return self.cross_entropy(
            scores,
            torch.zeros(
                scores.shape[0],
                dtype=torch.long,
                device=scores.device,
            ),
        )



import os
import random

import tqdm

__all__ = ["batchify", "iter"]


def batchify(
    X: list[str], batch_size: int, desc: str = "", tqdm_bar: bool = True
) -> list:
    os.environ["TOKENIZERS_PARALLELISM"] = "false"
    batchs = [X[pos : pos + batch_size] for pos in range(0, len(X), batch_size)]

    if tqdm_bar:
        for batch in tqdm.tqdm(
            batchs,
            position=0,
            total=1 + len(X) // batch_size,
            desc=desc,
        ):
            yield batch
    else:
        yield from batchs


def iter(
    X: list[tuple[str, str, float]],
    epochs: int,
    batch_size: int,
    shuffle: bool = True,
) -> tuple[list[str], list[str], list[float]]:
    """Iterate over a list of tuples (query, document, score) by batch of size batch_size.

    Parameters
    ----------
    X
        List of tuples (query, document, score).
    epochs
        Number of epochs.
    batch_size
        Batch size.
    shuffle
        Shuffle the data.

    Examples
    --------
    >>> from neural_cherche import utils

    >>> X = [
    ...    ("Apple", "🍏", "cherry"),
    ...    ("Banana", "🍌", "cherry"),
    ... ]

    >>> for anchor, positive, negative in utils.iter(
    ...         X,
    ...         epochs=1,
    ...         batch_size=3,
    ...         shuffle=False
    ...     ):
    ...     break

    >>> print(anchor)
    ['Apple', 'Banana']

    >>> print(positive)
    ['🍏', '🍌']

    >>> print(negative)
    ['cherry', 'cherry']

    """
    for epoch in range(epochs):
        if shuffle:
            random.shuffle(X)
        for batch in batchify(X=X, batch_size=batch_size, desc=f"Epoch {epoch}"):
            yield (
                [sample[0] for sample in batch],
                [sample[1] for sample in batch],
                [sample[2] for sample in batch],
            )



import numpy as np
import torch

from .. import models, utils
from ..retrieve.sparse_embed import SparseEmbed as SparseEmbedRetriever

__all__ = ["SparseEmbed"]


class SparseEmbed(SparseEmbedRetriever):
    """SparseEmbed ranker.

    Parameters
    ----------
    key
        Document unique identifier.
    on
        Document texts.
    model
        SparseEmbed model.
    device
        Device to use, default is model device.

    Examples
    --------
    >>> from neural_cherche import models, rank
    >>> from pprint import pprint
    >>> import torch

    >>> _ = torch.manual_seed(42)

    >>> model = models.SparseEmbed(
    ...     model_name_or_path="raphaelsty/neural-cherche-sparse-embed",
    ...     device="cpu",
    ...     embedding_size=64,
    ... )

    >>> ranker = rank.SparseEmbed(
    ...     key="id",
    ...     on="document",
    ...     model=model,
    ... )

    >>> documents = [
    ...     {"id": 0, "document": " ".join(["Food Hello world"] * 200)},
    ...     {"id": 1, "document": "Sports"},
    ...     {"id": 2, "document": "Cinema"},
    ... ]

    >>> queries = ["Food", "Sports", "Cinema"]

    >>> documents_embeddings = ranker.encode_documents(
    ...     documents=documents,
    ...     batch_size=2,
    ... )

    >>> queries_embeddings = ranker.encode_queries(
    ...     queries=queries,
    ...     batch_size=2,
    ... )

    >>> scores = ranker(
    ...     documents=[documents for _ in queries],
    ...     queries_embeddings=queries_embeddings,
    ...     documents_embeddings=documents_embeddings,
    ...     k=3,
    ...     batch_size=32
    ... )

    >>> pprint(scores)
    [[{'id': 0, 'similarity': 650.8815},
      {'id': 2, 'similarity': 63.361893},
      {'id': 1, 'similarity': 55.061672}],
     [{'id': 1, 'similarity': 139.74667},
      {'id': 2, 'similarity': 63.88109},
      {'id': 0, 'similarity': 42.964497}],
     [{'id': 2, 'similarity': 158.86716},
      {'id': 0, 'similarity': 68.95827},
      {'id': 1, 'similarity': 63.872555}]]

    >>> scores = ranker(
    ...     documents=[documents, [], documents],
    ...     queries_embeddings=queries_embeddings,
    ...     documents_embeddings=documents_embeddings,
    ...     k=3,
    ...     batch_size=32
    ... )

    >>> pprint(scores)
    [[{'id': 0, 'similarity': 650.8815},
      {'id': 2, 'similarity': 63.361893},
      {'id': 1, 'similarity': 55.061672}],
     [],
     [{'id': 2, 'similarity': 158.86716},
      {'id': 0, 'similarity': 68.95827},
      {'id': 1, 'similarity': 63.872555}]]

    >>> documents_embeddings = ranker.encode_documents(
    ...     documents=[documents for _ in queries],
    ...     batch_size=3,
    ... )

    >>> assert len(documents_embeddings) == len(documents)

    >>> documents_embeddings = ranker.encode_candidates_documents(
    ...     documents=documents,
    ...     candidates=[[{"id": 0}, {"id": 1}], [{"id": 1}], []],
    ...     batch_size=3,
    ... )

    >>> assert len(documents_embeddings) == 2

    """

    def __init__(
        self,
        key: str,
        on: list[str],
        model: models.SparseEmbed,
        tokenizer_parallelism: str = "false",
        device: str = None,
    ) -> None:
        super().__init__(
            key=key,
            on=on,
            model=model,
            tokenizer_parallelism=tokenizer_parallelism,
            device=device,
        )

    def encode_documents(
        self,
        documents: list[dict] | list[list[dict]],
        batch_size: int = 32,
        tqdm_bar: bool = True,
        query_mode: bool = False,
        desc: str = "documents embeddings",
        **kwargs,
    ) -> dict[str, dict[str, torch.Tensor]]:
        """Encode documents.

        Parameters
        ----------
        documents
            Documents to encode.
        batch_size
            Batch size.
        tqdm_bar
            Display a tqdm bar.
        query_mode
            Query mode. True if documents are queries.
        """
        if not documents:
            return {}

        # Flatten documents if necessary
        if isinstance(documents[0], list):
            documents_flatten, duplicates = [], {}
            for query_documents in documents:
                for document in query_documents:
                    if document[self.key] not in duplicates:
                        duplicates[document[self.key]] = True
                        documents_flatten.append(document)
            documents = documents_flatten

        return super().encode_documents(
            documents=documents,
            batch_size=batch_size,
            tqdm_bar=tqdm_bar,
            query_mode=query_mode,
            desc=desc,
            **kwargs,
        )

    def encode_candidates_documents(
        self,
        documents: list[dict],
        candidates: list[list[dict]],
        batch_size: int = 32,
        tqdm_bar: bool = True,
        query_mode: bool = False,
        **kwargs,
    ) -> dict[str, torch.Tensor]:
        """Map documents contents to candidates and encode them.
        This method is useful when you have a list of candidates for each query without
        their contents and you want to encode them.

        Parameters
        ----------
        documents
            Documents.
        candidates
            List of candidates for each query.
        batch_size
            Batch size.
        tqdm_bar
            Show tqdm bar.
        query_mode
            Query mode.
        """
        mapping_documents = {document[self.key]: document for document in documents}

        candidates = [
            {**candidate, **mapping_documents[candidate[self.key]]}
            for query_candidates in candidates
            for candidate in query_candidates
        ]

        return self.encode_documents(
            documents=candidates,
            batch_size=batch_size,
            tqdm_bar=tqdm_bar,
            query_mode=query_mode,
            **kwargs,
        )

    def __call__(
        self,
        documents: list[list[dict]],
        queries_embeddings: dict[str, torch.Tensor],
        documents_embeddings: dict[str, torch.Tensor],
        tqdm_bar: bool = True,
        k: int = None,
        batch_size: int = 32,
    ) -> list:
        """Retrieve documents.

        Parameters
        ----------
        documents
            Documents to rerank.
        queries_embeddings
            Queries embeddings.
        documents_embeddings
            Documents embeddings.
        tqdm_bar
            Display a tqdm bar.
        k
            Number of documents to retrieve.
        """
        k = k if k is not None else self.n_documents

        ranked = []

        for queries_batch, documents_batch in zip(
            utils.batchify(
                X=list(queries_embeddings.values()),
                batch_size=batch_size,
                desc=f"{self.__class__.__name__} ranker",
                tqdm_bar=tqdm_bar,
            ),
            utils.batchify(
                X=documents,
                batch_size=batch_size,
                tqdm_bar=False,
            ),
        ):
            queries_embeddings_batch = {
                "activations": torch.stack(
                    tensors=[
                        torch.tensor(
                            data=query["activations"],
                            device=self.model.device,
                            dtype=torch.int32,
                        )
                        for query in queries_batch
                    ],
                    dim=0,
                ),
                "embeddings": torch.stack(
                    tensors=[
                        torch.tensor(data=query["embeddings"], device=self.model.device)
                        for query in queries_batch
                    ],
                    dim=0,
                ),
            }

            documents_embeddings_batch = {}
            for query_documents in documents_batch:
                for document in query_documents:
                    documents_embeddings_batch[document[self.key]] = {
                        "activations": torch.tensor(
                            data=documents_embeddings[document[self.key]][
                                "activations"
                            ],
                            device=self.model.device,
                            dtype=torch.int32,
                        ),
                        "embeddings": documents_embeddings[document[self.key]][
                            "embeddings"
                        ],
                    }

            ranked.extend(
                self._rank(
                    queries_embeddings=queries_embeddings_batch,
                    documents_embeddings=documents_embeddings_batch,
                    documents=documents_batch,
                    k=k,
                )
            )

        return ranked

    def _rank(
        self,
        documents: list[list[dict]],
        queries_embeddings: dict[str, torch.Tensor],
        documents_embeddings: dict[str, torch.Tensor],
        k: int,
    ) -> list[list[dict]]:
        """Retrieve documents from input embeddings.

        Parameters
        ----------
        documents
            Documents to rerank.
        queries_embeddings
            Queries embeddings.
        documents_embeddings
            Documents embeddings.
        k
            Number of documents to retrieve.
        """
        documents_activations = [
            [
                documents_embeddings[document[self.key]]["activations"]
                for document in query_matchs
            ]
            for query_matchs in documents
        ]

        intersections = self._get_intersection(
            queries_activations=queries_embeddings["activations"],
            documents_activations=documents_activations,
        )

        queries_tokens_embeddings = [
            {
                token: token_embedding
                for token, token_embedding in zip(query_activations, query_embeddings)
            }
            for query_activations, query_embeddings in zip(
                queries_embeddings["activations"].cpu().detach().tolist(),
                queries_embeddings["embeddings"],
            )
        ]

        documents_tokens_embeddings = [
            [
                {
                    token: token_embedding
                    for token, token_embedding in zip(
                        document_activations.cpu().detach().tolist(),
                        documents_embeddings[document[self.key]]["embeddings"],
                    )
                }
                for document_activations, document in zip(
                    query_documents_activations, query_documents
                )
            ]
            for query_documents_activations, query_documents in zip(
                documents_activations, documents
            )
        ]

        dense_scores = self._get_scores(
            queries_embeddings=queries_tokens_embeddings,
            documents_embeddings=documents_tokens_embeddings,
            intersections=intersections,
        )

        ranked = []

        for query_scores, query_documents in zip(dense_scores, documents):
            if not query_documents:
                ranked.append([])
                continue

            query_scores, query_matchs = torch.topk(
                input=query_scores,
                k=min(k, query_scores.shape[0]),
                dim=-1,
            )

            query_matchs = np.take(
                a=query_documents, indices=query_matchs.cpu().detach().numpy()
            )

            query_scores = query_scores.cpu().detach().numpy()

            ranked.append(
                [
                    {self.key: document[self.key], "similarity": score}
                    for document, score in zip(query_matchs, query_scores)
                ]
            )

        return ranked



import torch

from .. import models, utils

__all__ = ["ColBERT"]


class ColBERT:
    """ColBERT ranker.

    Parameters
    ----------
    key
        Document unique identifier.
    on
        Document texts.
    model
        ColBERT model.
    device
        Device to use. default is model device.

    Examples
    --------
    >>> from neural_cherche import models, rank
    >>> from pprint import pprint
    >>> import torch

    >>> _ = torch.manual_seed(42)

    >>> encoder = models.ColBERT(
    ...     model_name_or_path="raphaelsty/neural-cherche-colbert",
    ...     device="mps",
    ... )

    >>> documents = [
    ...     {"id": 0, "document": "Food"},
    ...     {"id": 1, "document": "Sports"},
    ...     {"id": 2, "document": "Cinema"},
    ... ]

    >>> queries = ["Food", "Sports", "Cinema"]

    >>> ranker = rank.ColBERT(
    ...    key="id",
    ...    on=["document"],
    ...    model=encoder,
    ... )

    >>> queries_embeddings = ranker.encode_queries(
    ...     queries=queries,
    ...     batch_size=3,
    ... )

    >>> documents_embeddings = ranker.encode_documents(
    ...     documents=documents,
    ...     batch_size=3,
    ... )

    >>> scores = ranker(
    ...     documents=[documents for _ in queries],
    ...     queries_embeddings=queries_embeddings,
    ...     documents_embeddings=documents_embeddings,
    ...     batch_size=3,
    ...     tqdm_bar=True,
    ...     k=3,
    ... )

    >>> pprint(scores)
    [[{'document': 'Food', 'id': 0, 'similarity': 4.7243194580078125},
      {'document': 'Cinema', 'id': 2, 'similarity': 2.403003692626953},
      {'document': 'Sports', 'id': 1, 'similarity': 2.286036252975464}],
     [{'document': 'Sports', 'id': 1, 'similarity': 4.792296886444092},
      {'document': 'Cinema', 'id': 2, 'similarity': 2.6001152992248535},
      {'document': 'Food', 'id': 0, 'similarity': 2.312016487121582}],
     [{'document': 'Cinema', 'id': 2, 'similarity': 5.069696426391602},
      {'document': 'Sports', 'id': 1, 'similarity': 2.5587477684020996},
      {'document': 'Food', 'id': 0, 'similarity': 2.4474282264709473}]]

    >>> scores = ranker(
    ...     documents=[documents, [], documents],
    ...     queries_embeddings=queries_embeddings,
    ...     documents_embeddings=documents_embeddings,
    ...     batch_size=3,
    ...     tqdm_bar=True,
    ...     k=3,
    ... )

    >>> pprint(scores)
    [[{'document': 'Food', 'id': 0, 'similarity': 4.7243194580078125},
      {'document': 'Cinema', 'id': 2, 'similarity': 2.403003692626953},
      {'document': 'Sports', 'id': 1, 'similarity': 2.286036252975464}],
     [],
     [{'document': 'Cinema', 'id': 2, 'similarity': 5.069696426391602},
      {'document': 'Sports', 'id': 1, 'similarity': 2.5587477684020996},
      {'document': 'Food', 'id': 0, 'similarity': 2.4474282264709473}]]

    >>> documents_embeddings = ranker.encode_candidates_documents(
    ...     documents=documents,
    ...     candidates=[documents for _ in queries],
    ...     batch_size=3,
    ... )

    >>> assert len(documents_embeddings) == 3
    """

    def __init__(
        self,
        key: str,
        on: list[str],
        model: models.ColBERT,
        device: str = None,
    ) -> None:
        self.key = key
        self.on = on if isinstance(on, list) else [on]
        self.model = model
        self.device = self.model.device if device is None else device

    def encode_documents(
        self,
        documents: list[dict] | list[list[dict]],
        batch_size: int = 32,
        tqdm_bar: bool = True,
        query_mode: bool = False,
        desc: str = "documents embeddings",
        **kwargs,
    ) -> dict[str, torch.Tensor]:
        """Encode documents.

        Parameters
        ----------
        documents
            Documents.
        batch_size
            Batch size.
        tqdm_bar
            Show tqdm bar.
        """
        if not documents:
            return {}

        embeddings = self.encode_queries(
            queries=[
                " ".join([document[field] for field in self.on])
                for document in documents
            ],
            batch_size=batch_size,
            tqdm_bar=tqdm_bar,
            query_mode=query_mode,
            desc=desc,
            warn_duplicates=False,
            **kwargs,
        )

        return {
            document[self.key]: embedding
            for document, embedding in zip(documents, embeddings.values())
        }

    def encode_candidates_documents(
        self,
        documents: list[dict],
        candidates: list[list[dict]],
        batch_size: int = 32,
        tqdm_bar: bool = True,
        query_mode: bool = False,
        **kwargs,
    ) -> dict[str, torch.Tensor]:
        """Map documents contents to candidates and encode them.
        This method is useful when you have a list of candidates for each query without
        their contents and you want to encode them.

        Parameters
        ----------
        documents
            Documents.
        candidates
            List of candidates for each query.
        batch_size
            Batch size.
        tqdm_bar
            Show tqdm bar.
        query_mode
            Query mode.
        """
        mapping_documents = {document[self.key]: document for document in documents}

        documents_flatten, duplicates = [], {}
        for query_documents in candidates:
            for document in query_documents:
                if document[self.key] not in duplicates:
                    duplicates[document[self.key]] = True
                    documents_flatten.append(mapping_documents[document[self.key]])
        candidates = documents_flatten

        return self.encode_documents(
            documents=candidates,
            batch_size=batch_size,
            tqdm_bar=tqdm_bar,
            query_mode=query_mode,
            **kwargs,
        )

    def encode_queries(
        self,
        queries: list[str],
        batch_size: int = 32,
        tqdm_bar: bool = True,
        query_mode: bool = True,
        warn_duplicates: bool = True,
        desc: str = "queries embeddings",
        **kwargs,
    ) -> dict[str, torch.Tensor]:
        """Encode queries.

        Parameters
        ----------
        queries
            Queries.
        batch_size
            Batch size.
        tqdm_bar
            Show tqdm bar.
        """
        embeddings = {}

        for batch_texts in utils.batchify(
            X=queries,
            batch_size=batch_size,
            tqdm_bar=tqdm_bar,
            desc=f"{self.__class__.__name__} {desc}",
        ):
            batch_embeddings = self.model.encode(
                texts=batch_texts,
                query_mode=query_mode,
                **kwargs,
            )

            batch_embeddings = (
                batch_embeddings["embeddings"].cpu().detach().numpy().astype("float32")
            )

            for query, embedding in zip(batch_texts, batch_embeddings):
                embeddings[query] = embedding

        if len(embeddings) != len(queries) and warn_duplicates:
            utils.duplicates_queries_warning()

        return embeddings

    def __call__(
        self,
        documents: list[list[dict]],
        queries_embeddings: dict[str, torch.Tensor],
        documents_embeddings: dict[str, torch.Tensor],
        batch_size: int = 32,
        tqdm_bar: bool = True,
        k: int = None,
    ) -> list[list[str]]:
        """Rank documents  givent queries.

        Parameters
        ----------
        queries
            Queries.
        documents
            Documents.
        queries_embeddings
            Queries embeddings.
        documents_embeddings
            Documents embeddings.
        batch_size
            Batch size.
        tqdm_bar
            Show tqdm bar.
        k
            Number of documents to retrieve.
        """
        bar = utils.batchify(
            X=documents,
            batch_size=1,
            tqdm_bar=tqdm_bar,
            desc=f"{self.__class__.__name__} ranker",
        )

        scores = []

        for (query, query_embedding), query_documents in zip(
            queries_embeddings.items(), bar
        ):
            query_scores = []

            embedding_query = torch.tensor(
                data=query_embedding,
                device=self.device,
                dtype=torch.float32,
            )

            for batch_query_documents in utils.batchify(
                X=query_documents[0],
                batch_size=batch_size,
                tqdm_bar=False,
            ):
                embeddings_batch_documents = torch.stack(
                    tensors=[
                        torch.tensor(
                            data=documents_embeddings[document[self.key]],
                            device=self.device,
                            dtype=torch.float32,
                        )
                        for document in batch_query_documents
                    ],
                    dim=0,
                )

                query_documents_scores = torch.einsum(
                    "sh,bth->bst",
                    embedding_query,
                    embeddings_batch_documents,
                )

                query_scores.append(
                    query_documents_scores.max(dim=2).values.sum(axis=1)
                )

            if not query_scores:
                scores.append([])
                continue

            scores.append(torch.cat(tensors=query_scores, dim=0))

        return self._rank(scores=scores, documents=documents, k=k)

    def _rank(
        self, scores: torch.Tensor, documents: list[list[dict]], k: int
    ) -> list[list[dict]]:
        """Rank documents by scores.

        Parameters
        ----------
        scores
            Scores.
        documents
            Documents.
        k
            Number of documents to retrieve.
        """
        ranked = []

        for query_scores, query_documents in zip(scores, documents):
            if not query_documents:
                ranked.append([])
                continue

            top_k = torch.topk(
                input=query_scores,
                k=min(k, len(query_documents))
                if k is not None
                else len(query_documents),
                dim=-1,
            )

            ranked.append(
                [
                    {**query_documents[indice], "similarity": similarity}
                    for indice, similarity in zip(top_k.indices, top_k.values.tolist())
                ]
            )

        return ranked



import os
from abc import ABC, abstractmethod

import torch
from huggingface_hub import hf_hub_download
from transformers import AutoModelForMaskedLM, AutoTokenizer

from .. import utils


class Base(ABC, torch.nn.Module):
    """Base class from which all models inherit.

    Parameters
    ----------
    model_name_or_path
        Path to the model or the model name.
    device
        Device to use for the model. CPU or CUDA.
    kwargs
        Additional parameters to the model.
    """

    def __init__(
        self,
        model_name_or_path: str,
        device: str = None,
        extra_files_to_load: list[str] = [],
        query_prefix: str = "[Q] ",
        document_prefix: str = "[D] ",
        padding: str = "max_length",
        truncation: bool | None = True,
        add_special_tokens: bool = True,
        n_mask_tokens: int = 5,
        freeze_layers_except_last_n: int = None,
        **kwargs,
    ) -> None:
        """Initialize the model."""
        super(Base, self).__init__()

        self.query_prefix = query_prefix
        self.document_prefix = document_prefix
        self.padding = padding
        self.truncation = truncation
        self.add_special_tokens = add_special_tokens
        self.n_mask_tokens = n_mask_tokens

        if device is not None:
            self.device = device
        elif torch.cuda.is_available():
            self.device = "cuda"
        else:
            self.device = "cpu"

        os.environ["TRANSFORMERS_CACHE"] = "."
        self.model = AutoModelForMaskedLM.from_pretrained(
            pretrained_model_name_or_path=model_name_or_path, cache_dir="./", **kwargs
        ).to(self.device)

        # Download linear layer if exists
        for file in extra_files_to_load:
            try:
                _ = hf_hub_download(
                    repo_id=model_name_or_path, filename=file, cache_dir="."
                )
            except:
                pass

        self.tokenizer = AutoTokenizer.from_pretrained(
            pretrained_model_name_or_path=model_name_or_path,
            device=self.device,
            cache_dir="./",
            **kwargs,
        )

        self.model.config.output_hidden_states = True

        if os.path.exists(path=model_name_or_path):
            # Local checkpoint
            self.model_folder = model_name_or_path
        else:
            # HuggingFace checkpoint
            model_folder = os.path.join(
                f"models--{model_name_or_path}".replace("/", "--"), "snapshots"
            )
            snapshot = os.listdir(model_folder)[-1]
            self.model_folder = os.path.join(model_folder, snapshot)

        self.query_pad_token = self.tokenizer.mask_token
        self.original_pad_token = self.tokenizer.pad_token

        if freeze_layers_except_last_n is not None:
            self.model = utils.freeze_layers(
                model=self.model,
                n_layers=freeze_layers_except_last_n,
            )

    def _encode(self, texts: list[str], **kwargs) -> tuple[torch.Tensor, torch.Tensor]:
        """Encode sentences.

        Parameters
        ----------
        texts
            List of sentences to encode.
        """
        encoded_input = self.tokenizer.batch_encode_plus(
            batch_text_or_text_pairs=texts, return_tensors="pt", **kwargs
        )

        if self.device != "cpu":
            encoded_input = {
                key: value.to(self.device) for key, value in encoded_input.items()
            }

        output = self.model(**encoded_input)

        return (
            output.logits,
            output.hidden_states[-1],
            encoded_input["attention_mask"].unsqueeze(-1),
        )

    @abstractmethod
    def forward(self, *args, **kwargs):
        """Pytorch forward method."""
        pass

    @abstractmethod
    def encode(self, *args, **kwargs):
        """Encode documents."""
        pass

    @abstractmethod
    def scores(self, *args, **kwars):
        """Compute scores."""
        pass

    @abstractmethod
    def save_pretrained(self, path: str):
        """Save model the model."""
        pass



from .colbert_scores import colbert_scores
from .dense_scores import dense_scores
from .evaluate import evaluate, get_beir_triples, load_beir
from .freeze import freeze_layers
from .iter import batchify, iter
from .sparse_scores import sparse_scores
from .warnings import duplicates_queries_warning

__all__ = [
    "colbert_scores",
    "dense_scores",
    "freeze_layers",
    "get_beir_triples",
    "pairs_dense_scores",
    "evaluate",
    "load_beir",
    "batchify",
    "iter",
    "sparse_scores",
    "duplicates_queries_warning",
]



import setuptools

from neural_cherche.__version__ import __version__

with open(file="README.md", mode="r", encoding="utf-8") as fh:
    long_description = fh.read()

base_packages = [
    "torch >= 1.13",
    "tqdm >= 4.66",
    "transformers >= 4.34.0",
    "sentence-transformers >= 2.2.2",
    "lenlp >= 1.0.3",
]


eval = ["ranx >= 0.3.16", "beir >= 2.0.0"]

dev = [
    "mkdocs-material == 9.2.8",
    "mkdocs-awesome-pages-plugin == 2.9.2",
    " mkdocs-jupyter == 0.24.7",
]

setuptools.setup(
    name="neural_cherche",
    version=f"{__version__}",
    license="",
    author="Raphael Sourty",
    author_email="raphael.sourty@gmail.com",
    description="Sparse Embeddings for Neural Search.",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/raphaelsty/neural-cherche",
    download_url="https://github.com/user/neural-cherche/archive/v_01.tar.gz",
    keywords=[
        "neural search",
        "information retrieval",
        "semantic search",
        "SparseEmbed",
        "Google Research",
        "Splade",
        "Stanford",
        "ColBERT",
    ],
    packages=setuptools.find_packages(),
    install_requires=base_packages,
    extras_require={"eval": base_packages + eval, "dev": base_packages + eval + dev},
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    python_requires=">=3.6",
)



from transformers import AutoModelForMaskedLM

__all__ = ["freeze_layers"]


def freeze_layers(
    model: AutoModelForMaskedLM, n_layers: int = 5
) -> AutoModelForMaskedLM:
    """Freeze layers before the last n_layers."""
    trainable_layers = 0
    for name, param in model.named_parameters():
        if param.requires_grad:
            trainable_layers += 1

    for index, (name, param) in enumerate(iterable=model.named_parameters()):
        if index < (trainable_layers - 5):
            param.requires_grad = False

    return model



import json
import os

import torch

from .. import utils

__all__ = ["SparseEmbed"]

from .splade import Splade


class SparseEmbed(Splade):
    """SparseEmbed model.

    Parameters
    ----------
    model_name_or_path
        Path to the model or the model name. It should be a SentenceTransformer model.
    embedding_size
        Size of the embeddings in output of SparsEmbed model.
    kwargs
        Additional parameters to the pre-trained model.

    Examples
    --------
    >>> from neural_cherche import models
    >>> import torch

    >>> _ = torch.manual_seed(42)

    >>> model = models.SparseEmbed(
    ...     model_name_or_path="raphaelsty/neural-cherche-sparse-embed",
    ...     device="cpu",
    ... )

    >>> queries_embeddings = model.encode(
    ...     [" ".join(["Sports"]) * 500, "Music"],
    ... )

    >>> queries_embeddings["activations"].shape
    torch.Size([2, 64])

    >>> queries_embeddings["sparse_activations"].shape
    torch.Size([2, 30522])

    >>> queries_embeddings["embeddings"].shape
    torch.Size([2, 64, 128])

    >>> documents_embeddings = model.encode(
    ...    ["Music is great.", "Sports is great."],
    ...    query_mode=False,
    ... )

    >>> documents_embeddings["activations"].shape
    torch.Size([2, 256])

    >>> documents_embeddings["sparse_activations"].shape
    torch.Size([2, 30522])

    >>> documents_embeddings["embeddings"].shape
    torch.Size([2, 256, 128])

    >>> model.scores(
    ...     queries=["Sports", "Music"],
    ...     documents=["Sports is great.", "Music is great."],
    ...     batch_size=1,
    ... )
    tensor([101.4910, 196.2314])

    >>> _ = model.save_pretrained("checkpoint")

    >>> model = models.SparseEmbed(
    ...     model_name_or_path="checkpoint",
    ...     device="cpu",
    ... )

    >>> model.scores(
    ...     queries=["Sports", "Music"],
    ...     documents=["Sports is great.", "Music is great."],
    ...     batch_size=2,
    ... )
    tensor([101.4910, 196.2314])

    References
    ----------
    1. [SparseEmbed: Learning Sparse Lexical Representations with Contextual Embeddings for Retrieval](https://dl.acm.org/doi/pdf/10.1145/3539618.3592065)

    """

    def __init__(
        self,
        model_name_or_path: str = None,
        embedding_size: int = 128,
        max_length_query: int = 64,
        max_length_document: int = 256,
        device: str = None,
        query_prefix: str = "",
        document_prefix: str = "",
        padding: str = "max_length",
        truncation: bool | None = True,
        add_special_tokens: bool = True,
        n_mask_tokens: int = 3,
        freeze_layers_except_last_n: int = None,
        **kwargs,
    ) -> None:
        super(SparseEmbed, self).__init__(
            model_name_or_path=model_name_or_path,
            device=device,
            extra_files_to_load=["linear.pt", "metadata.json"],
            query_prefix=query_prefix,
            document_prefix=document_prefix,
            padding=padding,
            truncation=truncation,
            add_special_tokens=add_special_tokens,
            n_mask_tokens=n_mask_tokens,
            freeze_layers_except_last_n=freeze_layers_except_last_n,
            **kwargs,
        )

        self.embedding_size = embedding_size

        self.softmax = torch.nn.Softmax(dim=2).to(self.device)

        if os.path.exists(path=os.path.join(self.model_folder, "linear.pt")):
            linear = torch.load(
                f=os.path.join(self.model_folder, "linear.pt"), map_location=self.device
            )
            self.embedding_size = linear["weight"].shape[0]
            in_features = linear["weight"].shape[1]
        else:
            with torch.no_grad():
                _, embeddings, _ = self._encode(texts=["test"])
                in_features = embeddings.shape[2]

        self.linear = torch.nn.Linear(
            in_features=in_features,
            out_features=self.embedding_size,
            bias=False,
            device=self.device,
            dtype=torch.float32,
        )

        torch.nn.init.xavier_uniform_(
            tensor=self.linear.weight,
            gain=torch.nn.init.calculate_gain(nonlinearity="relu"),
        )

        if os.path.exists(path=os.path.join(self.model_folder, "linear.pt")):
            self.linear.load_state_dict(state_dict=linear)

        if os.path.exists(path=os.path.join(self.model_folder, "metadata.json")):
            with open(
                file=os.path.join(self.model_folder, "metadata.json"), mode="r"
            ) as file:
                metadata = json.load(fp=file)

            max_length_query = metadata["max_length_query"]
            max_length_document = metadata["max_length_document"]
            self.query_prefix = metadata.get("query_prefix", self.query_prefix)
            self.document_prefix = metadata.get("document_prefix", self.document_prefix)
            self.padding = metadata.get("padding", self.padding)
            self.truncation = metadata.get("truncation", self.truncation)
            self.add_special_tokens = metadata.get(
                "add_special_tokens", self.add_special_tokens
            )
            self.n_mask_tokens = metadata.get("n_mask_tokens", self.n_mask_tokens)

        self.max_length_query = max_length_query
        self.max_length_document = max_length_document

    def forward(
        self,
        texts: list[str],
        query_mode: bool = True,
        **kwargs,
    ) -> dict[str, torch.Tensor]:
        """Pytorch forward method.

        Parameters
        ----------
        texts
            List of documents to encode.
        query_mode
            Whether to encode queries or documents.
        """
        prefix = self.query_prefix if query_mode else self.document_prefix
        suffix = " ".join([self.tokenizer.mask_token] * self.n_mask_tokens)
        texts = [prefix + text + " " + suffix for text in texts]

        self.tokenizer.pad_token = (
            self.query_pad_token if query_mode else self.original_pad_token
        )

        k_tokens = self.max_length_query if query_mode else self.max_length_document

        logits, embeddings, attention_mask = self._encode(
            texts=texts,
            truncation=self.truncation,
            padding=self.padding,
            add_special_tokens=self.add_special_tokens,
            max_length=self.max_length_query
            if query_mode
            else self.max_length_document,
            **kwargs,
        )

        logits = logits * attention_mask
        embeddings = embeddings * attention_mask

        activations = self._update_activations(
            **self._get_activation(logits=logits),
            k_tokens=k_tokens,
        )

        attention = self._get_attention(
            logits=logits,
            activations=activations["activations"],
        )

        embeddings = torch.bmm(
            input=attention.transpose(dim0=1, dim1=2),
            mat2=embeddings,
        )

        return {
            "embeddings": self.relu(self.linear(embeddings)),
            "sparse_activations": activations["sparse_activations"],
            "activations": activations["activations"],
        }

    def _get_attention(
        self, logits: torch.Tensor, activations: torch.Tensor
    ) -> torch.Tensor:
        """Extract attention scores from MLM logits based on activated tokens."""
        attention = logits.gather(
            dim=2,
            index=torch.stack(
                tensors=[
                    torch.stack(tensors=[token for _ in range(logits.shape[1])])
                    for token in activations
                ]
            ),
        )

        return self.softmax(attention)

    def save_pretrained(self, path: str):
        """Save model the model."""
        self.model.save_pretrained(path)
        self.tokenizer.pad_token = self.original_pad_token
        self.tokenizer.save_pretrained(save_directory=path)
        torch.save(obj=self.linear.state_dict(), f=os.path.join(path, "linear.pt"))
        with open(file=os.path.join(path, "metadata.json"), mode="w") as file:
            json.dump(
                fp=file,
                obj={
                    "max_length_query": self.max_length_query,
                    "max_length_document": self.max_length_document,
                    "query_prefix": self.query_prefix,
                    "document_prefix": self.document_prefix,
                    "padding": self.padding,
                    "truncation": self.truncation,
                    "add_special_tokens": self.add_special_tokens,
                    "n_mask_tokens": self.n_mask_tokens,
                },
                indent=4,
            )

        return self

    def scores(
        self,
        queries: list[str],
        documents: list[str],
        batch_size: int = 32,
        tqdm_bar: bool = True,
        **kwargs,
    ) -> torch.Tensor:
        """Compute similarity scores between queries and documents."""
        dense_scores = []

        for batch_queries, batch_documents in zip(
            utils.batchify(
                X=queries,
                batch_size=batch_size,
                desc="Computing scores.",
                tqdm_bar=tqdm_bar,
            ),
            utils.batchify(X=documents, batch_size=batch_size, tqdm_bar=False),
        ):
            queries_embeddings = self.encode(
                texts=batch_queries,
                query_mode=True,
                **kwargs,
            )

            documents_embeddings = self.encode(
                texts=batch_documents,
                query_mode=False,
                **kwargs,
            )

            dense_scores.append(
                utils.pairs_dense_scores(
                    queries_activations=queries_embeddings["activations"],
                    documents_activations=documents_embeddings["activations"],
                    queries_embeddings=queries_embeddings["embeddings"],
                    documents_embeddings=documents_embeddings["embeddings"],
                )
            )

        return torch.cat(tensors=dense_scores, dim=0)
